from typing import TYPE_CHECKING

import pytest

from ruyi.ruyipkg.repo import (
    DEFAULT_REPO_ID,
    DEFAULT_REPO_NAME,
    DEFAULT_REPO_PRIORITY,
    RepoEntry,
)

if TYPE_CHECKING:
    from tests.fixtures import MockGlobalModeProvider
    from ruyi.log import RuyiLogger


class TestRepoEntry:
    def test_construction(self) -> None:
        entry = RepoEntry(
            id="my-vendor",
            name="My Vendor Overlay",
            remote="https://example.com/overlay.git",
            branch="main",
            local_path=None,
            priority=50,
            active=True,
        )
        assert entry.id == "my-vendor"
        assert entry.name == "My Vendor Overlay"
        assert entry.remote == "https://example.com/overlay.git"
        assert entry.branch == "main"
        assert entry.local_path is None
        assert entry.priority == 50
        assert entry.active is True

    def test_resolve_root_without_local_path(self) -> None:
        entry = RepoEntry(
            id="my-vendor",
            name="My Vendor Overlay",
            remote="https://example.com/overlay.git",
            branch="main",
            local_path=None,
            priority=50,
            active=True,
        )
        root = entry.resolve_root("/home/user/.cache/ruyi")
        assert root == "/home/user/.cache/ruyi/repos/my-vendor"

    def test_resolve_root_with_local_path(self) -> None:
        entry = RepoEntry(
            id="local-testing",
            name="Local Testing",
            remote=None,
            branch="main",
            local_path="/home/user/my-overlay",
            priority=100,
            active=True,
        )
        root = entry.resolve_root("/home/user/.cache/ruyi")
        assert root == "/home/user/my-overlay"

    def test_from_legacy_config(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        from ruyi.config import GlobalConfig

        gc = GlobalConfig(mock_gm, ruyi_logger)
        entry = RepoEntry.from_legacy_config(gc)

        assert entry.id == DEFAULT_REPO_ID
        assert entry.name == DEFAULT_REPO_NAME
        assert entry.priority == DEFAULT_REPO_PRIORITY
        assert entry.active is True
        assert entry.remote is not None
        assert entry.branch == "main"
        assert entry.local_path is None

    def test_from_legacy_config_with_overrides(
        self,
        mock_gm: "MockGlobalModeProvider",
        ruyi_logger: "RuyiLogger",
    ) -> None:
        from ruyi.config import GlobalConfig

        gc = GlobalConfig(mock_gm, ruyi_logger)
        gc.override_repo_url = "https://custom.example.com/repo.git"
        gc.override_repo_branch = "dev"
        gc.override_repo_dir = "/custom/path"

        entry = RepoEntry.from_legacy_config(gc)

        assert entry.remote == "https://custom.example.com/repo.git"
        assert entry.branch == "dev"
        assert entry.local_path == "/custom/path"

    def test_frozen(self) -> None:
        entry = RepoEntry(
            id="test",
            name="Test",
            remote=None,
            branch="main",
            local_path=None,
            priority=0,
            active=True,
        )
        with pytest.raises(AttributeError):
            entry.id = "changed"  # type: ignore[misc]



import io
import pathlib
import sys
from tests.fixtures import IntegrationTestHarness

import pytest

from ruyi.cli.main import is_version_query, main as ruyi_main
from ruyi.log import RuyiConsoleLogger
from ruyi.ruyipkg.distfile import Distfile

SHA_STUB = "0" * 64


class _TTYStringIO(io.StringIO):
    def isatty(self) -> bool:
        return True


def _fail_on_repo_access(self: object) -> None:
    raise AssertionError("completion setup must not access the package repo")


def test_cli_version_query_detection() -> None:
    assert is_version_query(["ruyi", "--version"])
    assert is_version_query(["ruyi", "-V"])
    assert is_version_query(["ruyi", "--porcelain", "--version"])
    assert is_version_query(["ruyi", "--config", "foo", "--version"])
    assert is_version_query(["ruyi", "version"])
    assert is_version_query(["ruyi", "--porcelain", "version"])
    assert not is_version_query(["ruyi"])
    assert not is_version_query(["ruyi", "list"])
    assert not is_version_query(["ruyi", "list", "version"])


def test_cli_version(ruyi_cli_runner: IntegrationTestHarness) -> None:
    for argv in [
        ["--version"],
        ["version"],
    ]:
        result = ruyi_cli_runner(*argv)
        assert result.exit_code == 0
        assert "Ruyi" in result.stdout
        assert "fatal error" not in result.stderr.lower()


def test_output_completion_script_does_not_access_repo_or_telemetry(
    ruyi_cli_runner: IntegrationTestHarness,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    from ruyi.ruyipkg.repo import MetadataRepo

    monkeypatch.setattr(MetadataRepo, "ensure_git_repo", _fail_on_repo_access)

    result = ruyi_cli_runner("--output-completion-script=bash")

    assert result.exit_code == 0
    assert result.stdout.startswith("#compdef ruyi\n")
    assert "package repository" not in result.stderr

    telemetry_root = pathlib.Path(ruyi_cli_runner._env["XDG_STATE_HOME"]) / "ruyi"
    assert not (telemetry_root / "telemetry" / "installation.json").exists()
    assert not (telemetry_root / "telemetry" / "minimal-installation-marker").exists()


def test_autocomplete_parser_build_does_not_access_package_repo(
    ruyi_cli_runner: IntegrationTestHarness,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    from ruyi.cli import builtin_commands
    from ruyi.cli.cmd import RootCommand
    from ruyi.ruyipkg.repo import MetadataRepo

    del builtin_commands

    ctx = ruyi_cli_runner.make_command_context()
    ctx.gm._is_cli_autocomplete = True  # pylint: disable=protected-access
    monkeypatch.setattr(MetadataRepo, "ensure_git_repo", _fail_on_repo_access)

    RootCommand.build_argparse(ctx.gc)


def test_cli_version_skips_first_run_oobe(
    ruyi_cli_runner: IntegrationTestHarness,
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    for argv in [
        ("--version",),
        ("version",),
    ]:
        ctx = ruyi_cli_runner.make_command_context(*argv)
        stdout = _TTYStringIO()
        stderr = _TTYStringIO()

        ctx.gc.logger = RuyiConsoleLogger(ctx.gm, stdout=stdout, stderr=stderr)
        monkeypatch.setattr(sys, "stdin", _TTYStringIO())
        monkeypatch.setattr(sys, "stdout", stdout)
        monkeypatch.setattr(sys, "stderr", stderr)

        exit_code = ruyi_main(ctx.gm, ctx.gc, ctx.argv)

        assert exit_code == 0
        assert "Ruyi" in stdout.getvalue()
        assert "Welcome to RuyiSDK" not in stderr.getvalue()

        telemetry_root = pathlib.Path(ctx.gc.telemetry_root)
        assert not (telemetry_root / "installation.json").exists()
        assert not (telemetry_root / "minimal-installation-marker").exists()


def test_cli_list_with_mock_repo(ruyi_cli_runner: IntegrationTestHarness) -> None:
    result = ruyi_cli_runner("list", "--name-contains", "sample-cli")

    assert result.exit_code == 0
    assert "dev-tools/sample-cli" in result.stdout


def test_cli_list_with_custom_package(ruyi_cli_runner: IntegrationTestHarness) -> None:
    manifest = (
        'format = "v1"\n'
        'kind = ["source"]\n\n'
        "[metadata]\n"
        'desc = "Custom integration package"\n'
        'vendor = { name = "Integration Tests", eula = "" }\n\n'
        "[[distfiles]]\n"
        'name = "custom-src.tar.zst"\n'
        "size = 0\n\n"
        "[distfiles.checksums]\n"
        f'sha256 = "{SHA_STUB}"\n'
    )
    ruyi_cli_runner.add_package("examples", "custom-cli", "0.1.0", manifest)

    result = ruyi_cli_runner("list", "--category-is", "examples")

    assert result.exit_code == 0
    assert "examples/custom-cli" in result.stdout


def test_cli_extract_without_subdir_reports_current_directory(
    ruyi_cli_runner: IntegrationTestHarness,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    manifest = f"""\
format = "v1"
kind = ["source"]

[metadata]
desc = "Extract message package"
vendor = {{ name = "Integration Tests", eula = "" }}

[[distfiles]]
name = "extract-src.raw"
size = 0
unpack = "raw"

[distfiles.checksums]
sha256 = "{SHA_STUB}"

[source]
distfiles = ["extract-src.raw"]
"""
    ruyi_cli_runner.add_package("examples", "extract-message", "1.0.0", manifest)

    unpack_roots: list[object] = []

    def fake_ensure(self: object, logger: object) -> None:
        pass

    def fake_unpack(self: object, root: object, logger: object) -> None:
        unpack_roots.append(root)

    monkeypatch.setattr(Distfile, "ensure", fake_ensure)
    monkeypatch.setattr(Distfile, "unpack", fake_unpack)

    result = ruyi_cli_runner(
        "extract",
        "--extract-without-subdir",
        "examples/extract-message",
    )

    assert result.exit_code == 0
    assert unpack_roots == [None]
    assert "has been extracted to ." in result.stderr
    assert "has been extracted to None" not in result.stderr



import pytest

from ruyi.log import RuyiLogger
from ruyi.pluginhost.ctx import PluginHostContext

from ..fixtures import RuyiFileFixtureFactory


@pytest.mark.xfail(
    reason="unsandboxed backend does not support freezing yet",
    strict=True,
)
def test_lang_frozen_values(
    ruyi_file: RuyiFileFixtureFactory,
    ruyi_logger: RuyiLogger,
) -> None:
    with ruyi_file.plugin_suite("lang_tests") as plugin_root:
        phctx = PluginHostContext.new(ruyi_logger, plugin_root)
        # Should fail because in the test plugin we're trying to append to a
        # frozen list.
        with pytest.raises(RuntimeError):
            phctx.get_from_plugin("frozen_values", "val")



"""Unsandboxed Python plugin host.

Rationale
---------

Ruyi plugins were originally authored in Starlark and executed through
``xingque`` (a binding to ``starlark-rust``), which provided genuine
language-level isolation. That backend was dropped in October 2024;
see commits ``bc458ce`` (the introduction of this file), ``a9d66bd``
(making it the default), and ``2b7c24d`` / ``9e5abe1`` (removal of the
Starlark backend).

The stated reason in ``bc458ce`` is that the plugin surface was
restricted to standard Python 3 rather than any dialect of it. The
underlying project-level goal is broader: the whole of RuyiSDK is kept
to Python and shell so that onboarding is trivial, so that loss of
project staff is survivable, and so that third-party commercial
partners can take over maintenance of their own forks without being
blocked by any non-trivial piece of code in a less widely known
language. Reintroducing Starlark -- or any other embedded language or
Rust-backed sandbox -- would reintroduce exactly the kind of
specialist-knowledge cliff this policy exists to avoid, and is
therefore not on the table regardless of its technical merits.

Threat model and non-goals
--------------------------

This module is intentionally *not* a sandbox. Plugin sources are
parsed, AST-linted, compiled, and ``exec``-ed in the host interpreter
with a curated ``__builtins__`` mapping. A malicious plugin can escape
trivially.

The original justification, as recorded in ``bc458ce`` in 2024, was:

    No "outsiders" are involved in plugin creation yet, and attacks
    from "insiders" are not going to be thwarted by code-level
    sandboxing alone.

That premise has since lapsed and the quote should be read as
historical rather than current. Third-party addon repositories are a
supported feature, which means plugin code loaded by Ruyi can now
originate from authors outside the project. An unsandboxed Python
runtime provides no meaningful defence against a hostile or
compromised third-party addon; in particular, the capability set in
``PluginHostContext`` (``call-subprocess-v1``, ``build-recipe-v1``,
``i18n-v1``, ...) is an API-shape boundary, not a security boundary,
and is trivially escapable from in-process Python.

The present mitigations against this are operational rather than
technical: a prominent warning and explicit user confirmation on
``ruyi repo add``, and the expectation that users extend trust to
third-party repos on the same footing as any other third-party code
they choose to run. Revisiting this with an actual sandbox (process
isolation, a reintroduced Starlark backend, or another mechanism) is
out of scope of this module and currently gated on project-level
decisions rather than technical ones; see
https://github.com/ruyisdk/ruyi/issues/444 for tracking.

Recursion detection, resource limits, timeouts, and filesystem or
network isolation are likewise out of scope here; they are not
soundly achievable with pure AST inspection plus ``exec`` in CPython.

What the module does enforce, and why, is documented on the
individual enforcement points themselves: ``BUILTINS_TO_EXPOSE``,
``GatedLanguageFeaturesPass``, and ``_load_stmt_helper`` below, plus
the ``PluginHostContext`` capability set in ``api.py`` /
``build_api.py``. The unifying goal of those checks is not security
but *Starlark portability*: keeping plugin sources close to the shape
of the original Starlark subset so that a future move back to a real
Starlark runtime -- should the project-level policy above ever shift
-- would not require rewriting every in-tree plugin.
"""

import ast
import builtins
import inspect
import os
import pathlib
from types import CodeType
from typing import Callable, Final, MutableMapping, NoReturn, TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Buffer

from .api import RuyiHostAPI
from .ctx import PluginHostContext, BasePluginLoader, PluginLoadMode


class UnsandboxedModuleDict(dict[str, object]):
    def get_option(self, key: str) -> object:
        return self.get(key, None)


class UnsandboxedTrivialEvaluator:
    def eval_function(
        self,
        function: object,
        *args: object,
        **kwargs: object,
    ) -> object:
        if callable(function):
            return function(*args, **kwargs)
        raise RuntimeError(f"the Python value {function!r} is not callable")


# The set of Python builtins exposed to plugin code in place of the real
# ``builtins`` module. Membership criterion: the name must either exist in
# Starlark or map cleanly onto a Starlark equivalent, so that keeping a
# plugin portable to a future Starlark backend does not require giving up
# any builtin listed here. Introspection / reflection / dynamic-eval
# builtins (``eval``, ``exec``, ``compile``, ``globals``, ``locals``,
# ``vars``, ``__import__``, ``open``, ...) are deliberately absent for the
# same reason: they have no Starlark counterpart. This is a portability
# fence, not a security boundary -- see the module docstring.
BUILTINS_TO_EXPOSE: Final = {
    k: getattr(builtins, k)
    for k in [
        "abs",
        "any",
        "all",
        "bool",
        "bytes",
        "dict",
        "dir",
        "enumerate",
        "float",
        "getattr",
        "hasattr",
        "hash",
        "int",
        "len",
        "list",
        "max",
        "min",
        "print",
        "range",
        "repr",
        "reversed",
        "sorted",
        "str",
        "tuple",
        "type",
        "zip",
    ]
}


def _fail_helper(*args: object) -> NoReturn:
    raise RuntimeError(f"fail: {''.join(str(x) for x in args)}")


class UnsandboxedPluginHostContext(
    PluginHostContext[UnsandboxedModuleDict, UnsandboxedTrivialEvaluator]
):
    def make_loader(
        self,
        originating_file: pathlib.Path,
        module_cache: MutableMapping[str, UnsandboxedModuleDict],
        load_mode: PluginLoadMode,
    ) -> BasePluginLoader[UnsandboxedModuleDict]:
        return UnsandboxedRuyiPluginLoader(
            self, originating_file, module_cache, load_mode
        )

    def make_evaluator(self) -> UnsandboxedTrivialEvaluator:
        return UnsandboxedTrivialEvaluator()


def _is_name_private(n: str) -> bool:
    return n.startswith("_")


def _assert_name_is_public(n: str) -> None | NoReturn:
    if _is_name_private(n):
        raise RuntimeError(f"error: trying to load private name {n}")
    return None


class UnsandboxedRuyiPluginLoader(BasePluginLoader[UnsandboxedModuleDict]):
    def do_load_module(
        self,
        resolved_path: pathlib.Path,
        program: str,
        ruyi_host_bridge: Callable[[object], RuyiHostAPI],
    ) -> UnsandboxedModuleDict:
        self.host_logger.D(f"unsandboxed module load: path {resolved_path}")

        sub_loader = self.make_sub_loader(resolved_path)

        def _load_stmt_helper(
            spec: str,
            *values_to_bind: str,
            **renamed_values_to_bind: str,
        ) -> None:
            """Starlark-style ``load()`` exposed to plugins.

            This is deliberately *not* a wrapper over Python ``import``:
            it mirrors Starlark's ``load()`` statement so that plugin
            sources remain portable to a real Starlark backend. In
            particular, it binds names by injection into the caller's
            frame (matching Starlark semantics), and it refuses to bind
            names beginning with ``_``, matching Starlark's rule that
            underscore-prefixed symbols are module-private.

            ``import`` / ``from ... import ...`` are separately rejected
            by ``GatedLanguageFeaturesPass`` so that ``load()`` is the
            only way for plugins to pull in other modules.
            """
            mod = sub_loader.load(spec)

            curr_frame = inspect.currentframe()
            if curr_frame is None:
                raise RuntimeError(
                    "cannot inspect the Python runtime for the current frame"
                )

            parent_frame = curr_frame.f_back
            if parent_frame is None:
                raise RuntimeError(
                    "internal error: no parent frame for load() statement"
                )

            g = parent_frame.f_locals
            for name in values_to_bind:
                _assert_name_is_public(name)
                g[name] = mod[name]
            for dst_name, src_name in renamed_values_to_bind.items():
                _assert_name_is_public(src_name)
                g[dst_name] = mod[src_name]
            return None

        code = self.source_to_code(program, resolved_path)
        mod_globals: dict[str, object] = {
            "__builtins__": BUILTINS_TO_EXPOSE,
            "fail": _fail_helper,
            "load": _load_stmt_helper,
            "ruyi_plugin_rev": ruyi_host_bridge,
        }
        # pylint: disable-next=exec-used
        exec(code, mod_globals)
        return UnsandboxedModuleDict(mod_globals)

    # intentionally follows the importlib.abc.InspectLoader protocol, for
    # easier refactoring whenever necessary.
    @staticmethod
    def source_to_code(
        data: "Buffer | str | ast.Module",
        path: "str | os.PathLike[str]" = "<string>",
    ) -> CodeType:
        mod_ast: ast.Module
        if isinstance(data, ast.Module):
            mod_ast = data
        else:  # isinstance(data, str) or isinstance(data, Buffer)
            mod_ast = ast.parse(data, path, "exec")

        # lint the module on a best-effort basis to help fight syntax feature
        # creep
        lint_module(mod_ast)

        return compile(mod_ast, path, "exec")


def lint_module(mod: ast.Module) -> None:
    """Run best-effort parse-time lints over a plugin module AST.

    Currently this runs only ``GatedLanguageFeaturesPass``; additional
    best-effort static checks (for example a call-graph pass flagging
    obvious direct or mutual recursion) may be layered in over time.
    Anything added here should be understood as a lint, not a soundness
    guarantee -- see the module docstring for why real enforcement is
    out of scope.
    """
    try:
        GatedLanguageFeaturesPass().visit(mod)
    except _GatedFeatureError as e:
        raise RuntimeError(
            f"line {e.node.lineno}: {e.feature} is not allowed in plugin code"
        ) from e


class _GatedFeatureError(Exception):
    """Internal signal raised by ``GatedLanguageFeaturesPass`` when it
    encounters a gated construct. Carries the offending AST node (for
    its line number) and a short human-readable name of the feature,
    so ``lint_module`` can surface a useful diagnostic instead of the
    bare node type.
    """

    def __init__(self, node: ast.stmt | ast.expr | ast.arg, feature: str) -> None:
        super().__init__(feature)
        self.node = node
        self.feature = feature


def _reject_annotated_args(args: ast.arguments) -> None:
    """Raise ``_GatedFeatureError`` if any parameter in ``args`` carries
    a type annotation. Starlark's Parameter grammar has no annotation
    production, so annotations on any category of parameter -- regular,
    positional-only, keyword-only, ``*args``, or ``**kwargs`` -- are
    rejected uniformly.
    """
    for arg in (
        *args.posonlyargs,
        *args.args,
        *args.kwonlyargs,
        *((args.vararg,) if args.vararg is not None else ()),
        *((args.kwarg,) if args.kwarg is not None else ()),
    ):
        if arg.annotation is not None:
            raise _GatedFeatureError(arg.annotation, "parameter type annotation")


def _find_slice_assign_target(target: ast.expr) -> ast.Subscript | None:
    """Return the offending ``Subscript`` node if ``target`` is, or
    directly contains within a top-level tuple/list unpacking, a
    subscript whose slice is an ``ast.Slice``. Return ``None`` otherwise.

    The check is deliberately shallow: Starlark's spec says "Starlark
    does not allow a slice expression to be the target of an assignment,
    although it may appear as a subexpression in the target," and
    forms like ``a[b[c:d]] = x`` are explicitly legal. We therefore
    only inspect the target spine itself (tuple/list unpacking), not
    arbitrary sub-expressions underneath a Subscript's value.
    """
    if isinstance(target, ast.Subscript) and isinstance(target.slice, ast.Slice):
        return target
    if isinstance(target, (ast.Tuple, ast.List)):
        for elt in target.elts:
            if (found := _find_slice_assign_target(elt)) is not None:
                return found
    return None


def _find_starred_in_target(target: ast.expr) -> ast.Starred | None:
    """Return the offending ``Starred`` node if ``target`` is, or
    contains anywhere within a tuple/list unpacking spine, a starred
    expression. Return ``None`` otherwise.

    Starlark's ``LoopVariables`` and ``AssignStmt`` LHS grammars do not
    permit ``*x`` unpacking; a starred target would have no portable
    equivalent. Starred expressions on the *right-hand side* or inside
    call arguments (``f(*args)``) are a separate grammatical position
    and are not checked here.
    """
    if isinstance(target, ast.Starred):
        return target
    if isinstance(target, (ast.Tuple, ast.List)):
        for elt in target.elts:
            if (found := _find_starred_in_target(elt)) is not None:
                return found
    return None


class GatedLanguageFeaturesPass(ast.NodeVisitor):
    """Reject Python syntax that has no Starlark analogue.

    Each ``visit_*`` override below names one construct that is gated
    at parse time. The selection criterion is *Starlark portability*,
    not safety: a feature is gated if accepting it would make it
    materially harder to move plugin sources back to a real Starlark
    runtime later. Rejected constructs include, among others:

    * ``NamedExpr`` (walrus) -- not in Starlark.
    * Decorators, return type annotations, parameter annotations and
      positional-only parameters on ``FunctionDef`` -- none of these
      have a production in Starlark's ``DefStmt`` or Parameter grammar.
    * ``AnnAssign`` (variable type annotations) -- no analogue.
    * ``JoinedStr`` (f-strings) -- Starlark has only plain string and
      bytes literals.
    * ``Set`` / ``SetComp`` -- Starlark has no set type or set
      comprehensions.
    * ``GeneratorExp`` -- Starlark has no generators.
    * ``Delete`` (``del`` statement) -- not in Starlark.
    * ``BinOp`` / ``AugAssign`` with ``MatMult`` (``@``, ``@=``) --
      not in Starlark's operator list.
    * ``Compare`` with more than one operator (chained comparisons) --
      Starlark's comparison operators are non-associative.
    * ``While`` -- not in Starlark's Statement grammar.
    * Slice expressions in assignment targets, and starred expressions
      in assignment / loop-variable targets -- forbidden by Starlark's
      ``AssignStmt`` and ``LoopVariables`` grammars respectively.
    * ``Raise``, ``Assert`` -- Starlark uses ``fail()`` for errors.
    * ``Import`` / ``ImportFrom`` -- Starlark uses ``load()``; see
      ``_load_stmt_helper``.
    * ``Try`` / ``TryStar`` / ``With`` -- no Starlark equivalents.
    * ``Match`` -- not in Starlark.
    * ``Yield`` / ``YieldFrom`` -- Starlark has no generators.
    * ``Global`` / ``Nonlocal`` -- Starlark's scoping rules differ.
    * ``ClassDef`` -- Starlark has no user-defined classes.
    * ``AsyncFunctionDef`` / ``Await`` / ``AsyncFor`` / ``AsyncWith``
      -- Starlark has no async model.

    The gate is necessarily best-effort: CPython's grammar is larger
    than Starlark's and evolves between releases. When a new language
    feature lands in CPython, the default choice should be to add it
    here -- anything not already required by an existing plugin is
    cheaper to forbid now than to un-ship later.

    This is a portability fence, not a security boundary; see the
    module docstring.
    """

    def visit_NamedExpr(self, node: ast.NamedExpr) -> None:
        raise _GatedFeatureError(node, "walrus operator (`:=`)")

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        # Decorators have no analogue in the Starlark ``DefStmt`` grammar.
        # Reject them, but continue recursing into the body so that other
        # gated constructs inside the function are still reported.
        if node.decorator_list:
            raise _GatedFeatureError(node.decorator_list[0], "decorator")
        # Starlark's parameter grammar has no annotation syntax, and its
        # function headers have no return-type annotation. Reject both.
        if node.returns is not None:
            raise _GatedFeatureError(node.returns, "return type annotation")
        _reject_annotated_args(node.args)
        # Starlark's Parameter grammar has no ``/`` marker (PEP 570), so
        # positional-only parameters have no way to be expressed there.
        if node.args.posonlyargs:
            raise _GatedFeatureError(
                node.args.posonlyargs[0], "positional-only parameter (`/`)"
            )
        self.generic_visit(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        raise _GatedFeatureError(node, "variable type annotation")

    def visit_JoinedStr(self, node: ast.JoinedStr) -> None:
        # Starlark's lexical grammar offers only plain string and bytes
        # literals. Reject f-strings so that plugin sources stay portable.
        raise _GatedFeatureError(node, "f-string")

    def visit_Set(self, node: ast.Set) -> None:
        # Starlark has no set type and therefore no set-display syntax.
        raise _GatedFeatureError(node, "set display")

    def visit_SetComp(self, node: ast.SetComp) -> None:
        # Starlark's comprehension grammar offers only ListComp and
        # DictComp; set comprehensions have no equivalent.
        raise _GatedFeatureError(node, "set comprehension")

    def visit_GeneratorExp(self, node: ast.GeneratorExp) -> None:
        # Starlark has no generators and no generator-expression
        # production in its grammar.
        raise _GatedFeatureError(node, "generator expression")

    def visit_Delete(self, node: ast.Delete) -> None:
        # Starlark's spec explicitly states that it does not have a
        # ``del`` statement.
        raise _GatedFeatureError(node, "`del` statement")

    def visit_While(self, node: ast.While) -> None:
        # Starlark's Statement grammar only has DefStmt, IfStmt, ForStmt
        # and SimpleStmt; ``while`` is listed among the reserved-but-
        # unused keywords. Rejecting it here also keeps plugin loops
        # bounded in the same way Starlark intends.
        raise _GatedFeatureError(node, "`while` loop")

    def visit_BinOp(self, node: ast.BinOp) -> None:
        # ``@`` (matrix multiplication) is not in Starlark's binary
        # operator list. Other binary operators are fine.
        if isinstance(node.op, ast.MatMult):
            raise _GatedFeatureError(node, "matrix-multiplication operator (`@`)")
        self.generic_visit(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        # Same reasoning for the in-place form ``@=``.
        if isinstance(node.op, ast.MatMult):
            raise _GatedFeatureError(node, "matrix-multiplication assignment (`@=`)")
        # Starlark forbids a slice expression on the LHS of an assignment.
        if (bad := _find_slice_assign_target(node.target)) is not None:
            raise _GatedFeatureError(bad, "slice as assignment target")
        # Starlark does not permit ``*x`` unpacking in an assignment LHS.
        if (starred := _find_starred_in_target(node.target)) is not None:
            raise _GatedFeatureError(starred, "starred assignment target")
        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        # Starlark forbids a slice expression on the LHS of an assignment.
        for target in node.targets:
            if (bad := _find_slice_assign_target(target)) is not None:
                raise _GatedFeatureError(bad, "slice as assignment target")
            if (starred := _find_starred_in_target(target)) is not None:
                raise _GatedFeatureError(starred, "starred assignment target")
        self.generic_visit(node)

    def visit_For(self, node: ast.For) -> None:
        # Starlark's LoopVariables grammar permits no ``*x`` unpacking.
        if (starred := _find_starred_in_target(node.target)) is not None:
            raise _GatedFeatureError(starred, "starred loop variable")
        self.generic_visit(node)

    def visit_ListComp(self, node: ast.ListComp) -> None:
        for gen in node.generators:
            if (starred := _find_starred_in_target(gen.target)) is not None:
                raise _GatedFeatureError(starred, "starred loop variable")
        self.generic_visit(node)

    def visit_DictComp(self, node: ast.DictComp) -> None:
        for gen in node.generators:
            if (starred := _find_starred_in_target(gen.target)) is not None:
                raise _GatedFeatureError(starred, "starred loop variable")
        self.generic_visit(node)

    def visit_Compare(self, node: ast.Compare) -> None:
        # Starlark spec: "Comparison operators, ``in``, and ``not in`` are
        # non-associative, so the parser will not accept ``0 <= i < n``."
        # Reject chained comparisons (more than one operator in a single
        # Compare node) accordingly.
        if len(node.ops) > 1:
            raise _GatedFeatureError(node, "chained comparison")
        # Starlark's comparison operator list is ``==``, ``!=``, ``<``,
        # ``>``, ``<=``, ``>=``, ``in`` and ``not in``; there are no
        # identity operators.
        for op in node.ops:
            if isinstance(op, ast.Is):
                raise _GatedFeatureError(node, "`is` operator")
            if isinstance(op, ast.IsNot):
                raise _GatedFeatureError(node, "`is not` operator")
        self.generic_visit(node)

    def visit_Raise(self, node: ast.Raise) -> None:
        raise _GatedFeatureError(node, "`raise` statement")

    def visit_Assert(self, node: ast.Assert) -> None:
        raise _GatedFeatureError(node, "`assert` statement")

    def visit_Import(self, node: ast.Import) -> None:
        raise _GatedFeatureError(node, "`import` statement")

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        raise _GatedFeatureError(node, "`from ... import ...` statement")

    def visit_Try(self, node: ast.Try) -> None:
        raise _GatedFeatureError(node, "`try` statement")

    def visit_TryStar(self, node: ast.TryStar) -> None:
        raise _GatedFeatureError(node, "`try ... except*` statement")

    def visit_With(self, node: ast.With) -> None:
        raise _GatedFeatureError(node, "`with` statement")

    def visit_Match(self, node: ast.Match) -> None:
        raise _GatedFeatureError(node, "`match` statement")

    def visit_Yield(self, node: ast.Yield) -> None:
        raise _GatedFeatureError(node, "`yield` expression")

    def visit_YieldFrom(self, node: ast.YieldFrom) -> None:
        raise _GatedFeatureError(node, "`yield from` expression")

    def visit_Global(self, node: ast.Global) -> None:
        raise _GatedFeatureError(node, "`global` statement")

    def visit_Nonlocal(self, node: ast.Nonlocal) -> None:
        raise _GatedFeatureError(node, "`nonlocal` statement")

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        raise _GatedFeatureError(node, "`class` definition")

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        raise _GatedFeatureError(node, "`async def` function")

    def visit_Await(self, node: ast.Await) -> None:
        raise _GatedFeatureError(node, "`await` expression")

    def visit_AsyncFor(self, node: ast.AsyncFor) -> None:
        raise _GatedFeatureError(node, "`async for` loop")

    def visit_AsyncWith(self, node: ast.AsyncWith) -> None:
        raise _GatedFeatureError(node, "`async with` statement")



import functools
import re
from typing import Any, Final, TypedDict

from ..config.news import NewsReadStatusStore
from ..utils import frontmatter
from ..utils.porcelain import PorcelainEntity, PorcelainEntityType
from ..utils.l10n import match_lang_code

NEWS_FILENAME_RE: Final = re.compile(r"^(\d+-\d{2}-\d{2}-.*?)(\.[0-9A-Za-z_-]+)?\.md$")


@functools.total_ordering
class NewsItemNameMetadata:
    def __init__(self, id: str, lang: str) -> None:
        self.id = id
        self.lang = lang

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, NewsItemNameMetadata):
            return NotImplemented
        return self.id == other.id

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, NewsItemNameMetadata):
            return NotImplemented

        # order by id in lexical order
        return self.id < other.id


def parse_news_filename(filename: str) -> NewsItemNameMetadata | None:
    m = NEWS_FILENAME_RE.match(filename)
    if m is None:
        return None

    id = m.group(1)
    lang = m.group(2)
    if not lang:
        lang = "zh_CN"  # TODO: kill after l10n work is complete
    else:
        lang = lang[1:]  # strip the dot prefix

    return NewsItemNameMetadata(id, lang)


class NewsItemStore:
    def __init__(self, rs_store: NewsReadStatusStore) -> None:
        self._buf_news_by_ids: dict[str, NewsItem] = {}
        self._newsitems: list[NewsItem]
        self._rs_store = rs_store

    def add(self, filename: str, contents: str) -> None:
        md = parse_news_filename(filename)
        if md is None:
            return None

        post = frontmatter.loads(contents)

        if ni := self._buf_news_by_ids.get(md.id):
            ni.add_lang(md, post)
        else:
            ni = NewsItem(md.id)
            ni.add_lang(md, post)
            self._buf_news_by_ids[md.id] = ni

    def add_item(
        self,
        news_id: str,
        md: NewsItemNameMetadata,
        post: frontmatter.Post,
    ) -> None:
        """Add a pre-parsed news item entry, used for cross-repo merging."""
        if ni := self._buf_news_by_ids.get(news_id):
            ni.add_lang(md, post)
        else:
            ni = NewsItem(news_id)
            ni.add_lang(md, post)
            self._buf_news_by_ids[news_id] = ni

    def finalize(self) -> None:
        self._newsitems = list(self._buf_news_by_ids.values())
        # sort in intended display order
        self._newsitems.sort()

        # mark the news item instances with ordinals
        for i, ni in enumerate(self._newsitems):
            ni.ordinal = i + 1

        # also read statuses
        for ni in self._newsitems:
            ni.is_read = ni.id in self._rs_store

    def list(self, only_unread: bool) -> list["NewsItem"]:
        if not only_unread:
            return self._newsitems
        return [x for x in self._newsitems if not x.is_read]

    def mark_as_read(self, *ids: str) -> None:
        if not ids:
            return
        for id in ids:
            self._rs_store.add(id)
        self._rs_store.save()


@functools.total_ordering
class NewsItem:
    def __init__(self, id: str) -> None:
        self._id = id
        self._content_by_lang: dict[str, NewsItemContent] = {}

        # these fields are updated later in store initialization code
        self.ordinal = 0
        self.is_read = False

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, NewsItem):
            return NotImplemented
        return self._id == other._id and self.ordinal == other.ordinal

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, NewsItem):
            return NotImplemented
        return self._id < other._id or self.ordinal < other.ordinal

    @property
    def id(self) -> str:
        return self._id

    @property
    def langs(self) -> dict[str, "NewsItemContent"]:
        """Language-keyed content map."""
        return self._content_by_lang

    def __contains__(self, lang: str) -> bool:
        return lang in self._content_by_lang

    def __getitem__(self, lang: str) -> "NewsItemContent":
        return self._content_by_lang[lang]

    def add_lang(self, md: NewsItemNameMetadata, post: frontmatter.Post) -> None:
        self._content_by_lang[md.lang] = NewsItemContent(md, post)

    def __delitem__(self, lang: str) -> None:
        del self._content_by_lang[lang]

    def get_content_for_lang(self, lang: str) -> "NewsItemContent":
        resolved_lang_code = match_lang_code(lang, self._content_by_lang.keys())
        return self[resolved_lang_code]

    def to_porcelain(self) -> "PorcelainNewsItemV1":
        return {
            "ty": PorcelainEntityType.NewsItemV1,
            "id": self.id,
            "ord": self.ordinal,
            "is_read": self.is_read,
            "langs": [x.to_porcelain() for x in self._content_by_lang.values()],
        }


class NewsItemContent:
    def __init__(
        self,
        md: NewsItemNameMetadata,
        post: frontmatter.Post,
    ) -> None:
        self._md = md
        self._post = post

    @property
    def lang(self) -> str:
        return self._md.lang

    @property
    def post(self) -> frontmatter.Post:
        """The parsed frontmatter post."""
        return self._post

    @property
    def metadata(self) -> NewsItemNameMetadata:
        """The filename-derived metadata."""
        return self._md

    @property
    def display_title(self) -> str:
        metadata_title = self._post.get("title")
        return metadata_title if isinstance(metadata_title, str) else self._md.id

    @property
    def content(self) -> str:
        return self._post.content

    def to_porcelain(self) -> "PorcelainNewsItemContentV1":
        return {
            "lang": self.lang,
            "display_title": self.display_title,
            "content": self.content,
        }


class PorcelainNewsItemContentV1(TypedDict):
    lang: str
    display_title: str
    content: str


class PorcelainNewsItemV1(PorcelainEntity):
    id: str
    ord: int
    is_read: bool
    langs: list[PorcelainNewsItemContentV1]



from typing import Iterable, Protocol

from .pkg_manifest import BoundPackageManifest


class ProvidesPackageManifests(Protocol):
    """A protocol that defines methods for providing package manifests."""

    def get_pkg(
        self,
        name: str,
        category: str,
        ver: str,
    ) -> BoundPackageManifest | None:
        """Returns the package manifest by exact match, or None if not found."""
        ...

    def iter_pkg_manifests(self) -> Iterable[BoundPackageManifest]:
        """Iterates over all package manifests provided by this store."""
        ...

    def iter_pkgs(
        self,
    ) -> Iterable[tuple[str, str, dict[str, BoundPackageManifest]]]:
        """Iterates over all package manifests provided by this store, returning
        ``(category, package_name, pkg_manifests_by_versions)``."""
        ...

    def iter_pkg_vers(
        self,
        name: str,
        category: str | None = None,
    ) -> Iterable[BoundPackageManifest]:
        """Iterates over all versions of a certain package provided by this store,
        specified by name and optionally category."""
        ...

    def get_pkg_latest_ver(
        self,
        name: str,
        category: str | None = None,
        include_prerelease_vers: bool = False,
    ) -> BoundPackageManifest:
        """Returns the latest version of a package provided by this store,
        specified by name and optionally category.

        If ``include_prerelease_vers`` is True, it will also consider prerelease
        versions. Raises KeyError if no such package exists."""
        ...

    # To be removed later along with slug support
    def get_pkg_by_slug(self, slug: str) -> BoundPackageManifest | None:
        """Returns the package with the specified slug from this store, or None
        if not found."""
        ...



#!/usr/bin/env python3

import os
import pathlib
import subprocess
import sys
from typing import NoReturn

from pygit2 import Repository
from tomlkit_extras import TOMLDocumentDescriptor, load_toml_file


try:
    from semver.version import Version  # type: ignore[import-untyped,unused-ignore]
except ModuleNotFoundError:
    from semver import VersionInfo as Version  # type: ignore[import-untyped,unused-ignore]


def render_tag_message(version: str) -> str:
    return f"Ruyi {version}"


def fatal(msg: str) -> NoReturn:
    print(f"fatal: {msg}", file=sys.stderr)
    sys.exit(1)


def main() -> None:
    # assume CWD is project root, which is guaranteed to be the case (see
    # end of file)
    pyproject = load_toml_file(pathlib.Path("pyproject.toml"))
    pyproject_desc = TOMLDocumentDescriptor(pyproject)

    project_table = pyproject_desc.get_table("project")
    version_field = project_table.fields["version"]
    lineno = version_field.line_no
    version = version_field.value
    if not isinstance(version, str):
        fatal(f"expected project.version to be a string, got {type(version)}")

    # Check if the version is a valid semver version
    try:
        Version.parse(version)
    except ValueError as e:
        fatal(f"invalid semver {version} in pyproject.toml: {e}")

    print(f"info: project version is {version}, defined at pyproject.toml:{lineno}")

    # Check if the tag is already present
    repo = Repository(".")
    try:
        tag_ref = repo.lookup_reference(f"refs/tags/{version}")
    except KeyError:
        tag_ref = None

    if tag_ref is not None:
        print(f"info: tag {version} already exists")
        # idempotence: don't fail the workflow with non-zero status code
        sys.exit(0)

    # Blame pyproject.toml to find the commit bumping the version
    blame = repo.blame("pyproject.toml")
    ver_bump_commit_id = blame.for_line(lineno).final_commit_id
    print(f"info: the version-bumping commit is {ver_bump_commit_id}")

    ver_bump_commit = repo.get(ver_bump_commit_id)
    if ver_bump_commit is None:
        fatal(f"could not find version-bumping commit {ver_bump_commit_id}")

    # Create the tag with Git command line to allow for GPG signing
    argv = ["git", "tag", "-m", render_tag_message(version)]

    if "RUYI_NO_GPG_SIGN" in os.environ:
        argv.extend(["-a", "--no-sign"])
    else:
        argv.append("-s")

    argv.extend([version, str(ver_bump_commit_id)])

    print(f"info: invoking git: {' '.join(argv)}")
    subprocess.run(argv, check=True)

    print(f"info: tag {version} created successfully")
    sys.exit(0)


if __name__ == "__main__":
    # cd to project root
    os.chdir(os.path.join(os.path.dirname(__file__), ".."))
    main()



import ctypes
import os
import ssl
import sys
from typing import Final, NamedTuple

import certifi

from ..i18n import _
from ..log import RuyiConsoleLogger, RuyiLogger
from .global_mode import EnvGlobalModeProvider

_orig_get_default_verify_paths: Final = ssl.get_default_verify_paths
_cached_paths: ssl.DefaultVerifyPaths | None = None


def get_system_ssl_default_verify_paths() -> ssl.DefaultVerifyPaths:
    global _cached_paths

    if _cached_paths is None:
        # no way to pass in the logger because of the function signature
        # so we have to use a new logger
        gm = EnvGlobalModeProvider(os.environ)
        _cached_paths = _get_system_ssl_default_verify_paths(RuyiConsoleLogger(gm))
    return _cached_paths


def _get_system_ssl_default_verify_paths(logger: RuyiLogger) -> ssl.DefaultVerifyPaths:
    orig_paths = _orig_get_default_verify_paths()
    if sys.platform != "linux":
        return orig_paths

    result: ssl.DefaultVerifyPaths | None = None

    # imitate the stdlib flow but with overridden data source
    try:
        parts = _query_linux_system_ssl_default_cert_paths(logger)
        if parts is None:
            logger.W(_("failed to probe system libcrypto"))
        else:
            result = to_ssl_paths(parts)
    except Exception as e:
        logger.D(f"cannot get system libcrypto default cert paths: {e}")

    if result is None:
        logger.D("falling back to probing hard-coded paths")
        result = probe_fallback_verify_paths()

    if result != orig_paths:
        logger.D(
            "get_default_verify_paths() values differ between bundled and system libssl"
        )
        logger.D(f"bundled: {orig_paths}")
        logger.D(f" system: {result}")

    return result


def to_ssl_paths(parts: tuple[str, str, str, str]) -> ssl.DefaultVerifyPaths | None:
    cafile = os.environ.get(parts[0], parts[1])
    capath = os.environ.get(parts[2], parts[3])

    is_cafile_present = os.path.isfile(cafile)
    is_capath_present = os.path.isdir(capath)
    if not is_cafile_present and not is_capath_present:
        return None

    # must do "else None" like the stdlib, despite the type annotation being just "str"
    return ssl.DefaultVerifyPaths(
        cafile if is_cafile_present else None,  # type: ignore[arg-type]
        capath if is_capath_present else None,  # type: ignore[arg-type]
        *parts,
    )


def _decode_fsdefault_or_none(val: int | None) -> str:
    if val is None:
        return ""

    s = ctypes.c_char_p(val)
    if s.value is None:
        return ""

    return s.value.decode(sys.getfilesystemencoding())


def _query_linux_system_ssl_default_cert_paths(
    logger: RuyiLogger,
    soname: str | None = None,
) -> tuple[str, str, str, str] | None:
    if soname is None:
        # check libcrypto instead of libssl, because if the system libssl is
        # newer than the bundled one, the system libssl will depend on the
        # bundled libcrypto that may lack newer ELF symbol version(s). The
        # functions actually reside in libcrypto, after all.
        for soname in ("libcrypto.so", "libcrypto.so.3", "libcrypto.so.1.1"):
            try:
                return _query_linux_system_ssl_default_cert_paths(logger, soname)
            except OSError as e:
                logger.D(f"soname {soname} not working: {e}")
                continue

        return None

    # dlopen-ing the bare soname will get us the system library
    lib = ctypes.CDLL(soname)
    lib.X509_get_default_cert_file_env.restype = ctypes.c_void_p
    lib.X509_get_default_cert_file.restype = ctypes.c_void_p
    lib.X509_get_default_cert_dir_env.restype = ctypes.c_void_p
    lib.X509_get_default_cert_dir.restype = ctypes.c_void_p

    result = (
        _decode_fsdefault_or_none(lib.X509_get_default_cert_file_env()),
        _decode_fsdefault_or_none(lib.X509_get_default_cert_file()),
        _decode_fsdefault_or_none(lib.X509_get_default_cert_dir_env()),
        _decode_fsdefault_or_none(lib.X509_get_default_cert_dir()),
    )

    logger.D(f"got defaults from system libcrypto {soname}")
    logger.D(f"X509_get_default_cert_file_env() = {result[0]}")
    logger.D(f"X509_get_default_cert_file() = {result[1]}")
    logger.D(f"X509_get_default_cert_dir_env() = {result[2]}")
    logger.D(f"X509_get_default_cert_dir() = {result[3]}")

    return result


class WellKnownCALocation(NamedTuple):
    cafile: str
    capath: str


WELL_KNOWN_CA_LOCATIONS: Final[list[WellKnownCALocation]] = [
    # Debian-based distros
    WellKnownCALocation("/usr/lib/ssl/cert.pem", "/usr/lib/ssl/certs"),
    # RPM-based distros
    WellKnownCALocation("/etc/pki/tls/cert.pem", "/etc/pki/tls/certs"),
    # Most others
    WellKnownCALocation("/etc/ssl/cert.pem", "/etc/ssl/certs"),
]


def probe_fallback_verify_paths() -> ssl.DefaultVerifyPaths:
    for loc in WELL_KNOWN_CA_LOCATIONS:
        is_file_present = os.path.isfile(loc.cafile)
        is_dir_present = os.path.isdir(loc.capath)
        if not is_file_present and not is_dir_present:
            continue

        return ssl.DefaultVerifyPaths(
            loc.cafile if is_file_present else None,  # type: ignore[arg-type]
            loc.capath if is_dir_present else None,  # type: ignore[arg-type]
            "SSL_CERT_FILE",
            loc.cafile,
            "SSL_CERT_DIR",
            loc.capath,
        )

    # fall back to certifi
    cafile = certifi.where()
    return ssl.DefaultVerifyPaths(
        cafile,
        None,  # type: ignore[arg-type]
        "SSL_CERT_FILE",
        cafile,
        "SSL_CERT_DIR",
        "/etc/ssl/certs",
    )


ssl.get_default_verify_paths = get_system_ssl_default_verify_paths



import argparse
from typing import TYPE_CHECKING

from ..cli.cmd import RootCommand
from ..i18n import _

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class DeviceCommand(
    RootCommand,
    cmd="device",
    has_subcommands=True,
    help=_("Manage devices"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass


class DeviceProvisionCommand(
    DeviceCommand,
    cmd="provision",
    aliases=["flash"],
    help=_("Interactively initialize a device for development"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .provision import do_provision_interactive

        try:
            return do_provision_interactive(cfg)
        except KeyboardInterrupt:
            cfg.logger.stdout(
                _("\n\nKeyboard interrupt received, exiting."), end="\n\n"
            )
            return 1



import os
import pathlib

import pytest

from ruyi.ruyipkg.recipe_project import (
    MARKER_FILENAME,
    RecipeProject,
    RecipeProjectError,
    discover_recipe_project,
    safe_join,
)


def _write_marker(root: pathlib.Path, body: str) -> None:
    (root / MARKER_FILENAME).write_text(body)


def test_discover_from_immediate_parent(tmp_path: pathlib.Path) -> None:
    _write_marker(tmp_path, 'format = "v1"\n[project]\nname = "demo"\n')
    recipe = tmp_path / "foo.star"
    recipe.write_text("")

    rp = discover_recipe_project(recipe)
    assert rp.root == tmp_path.resolve()
    assert rp.name == "demo"
    assert rp.output_dir == (tmp_path / "out").resolve()
    assert rp.extra_artifact_roots == ()


def test_discover_walks_up_parents(tmp_path: pathlib.Path) -> None:
    _write_marker(tmp_path, 'format = "v1"\n')
    nested = tmp_path / "a" / "b" / "c"
    nested.mkdir(parents=True)
    recipe = nested / "x.star"
    recipe.write_text("")

    rp = discover_recipe_project(recipe)
    assert rp.root == tmp_path.resolve()
    # Default name falls back to root dir name.
    assert rp.name == tmp_path.name


def test_discover_missing_marker(tmp_path: pathlib.Path) -> None:
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    with pytest.raises(RecipeProjectError, match="no ruyi-build-recipes.toml"):
        discover_recipe_project(recipe)


def test_discover_missing_recipe_file(tmp_path: pathlib.Path) -> None:
    with pytest.raises(RecipeProjectError, match="recipe file not found"):
        discover_recipe_project(tmp_path / "nope.star")


def test_discover_malformed_toml(tmp_path: pathlib.Path) -> None:
    _write_marker(tmp_path, "not = valid = toml")
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    with pytest.raises(RecipeProjectError, match="malformed"):
        discover_recipe_project(recipe)


def test_discover_unsupported_format(tmp_path: pathlib.Path) -> None:
    _write_marker(tmp_path, 'format = "v2"\n')
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    with pytest.raises(RecipeProjectError, match="unsupported or missing 'format'"):
        discover_recipe_project(recipe)


def test_discover_custom_output_dir_and_extras(tmp_path: pathlib.Path) -> None:
    _write_marker(
        tmp_path,
        'format = "v1"\n'
        "[project]\n"
        'name = "demo"\n'
        'output_dir = "dist"\n'
        f'extra_artifact_roots = ["{tmp_path.as_posix()}/scratch"]\n',
    )
    (tmp_path / "scratch").mkdir()
    recipe = tmp_path / "x.star"
    recipe.write_text("")

    rp = discover_recipe_project(recipe)
    assert rp.output_dir == (tmp_path / "dist").resolve()
    assert rp.extra_artifact_roots == ((tmp_path / "scratch").resolve(),)


def test_discover_absolute_output_dir_rejected(tmp_path: pathlib.Path) -> None:
    _write_marker(
        tmp_path,
        'format = "v1"\n[project]\noutput_dir = "/abs"\n',
    )
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    with pytest.raises(
        RecipeProjectError, match="output_dir must be a project-relative"
    ):
        discover_recipe_project(recipe)


def test_discover_traversing_output_dir_rejected(tmp_path: pathlib.Path) -> None:
    _write_marker(
        tmp_path,
        'format = "v1"\n[project]\noutput_dir = "../out"\n',
    )
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    with pytest.raises(RecipeProjectError, match="escapes the project root"):
        discover_recipe_project(recipe)


def test_discover_relative_extra_artifact_root_rejected(tmp_path: pathlib.Path) -> None:
    _write_marker(
        tmp_path,
        'format = "v1"\n[project]\nextra_artifact_roots = ["scratch"]\n',
    )
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    with pytest.raises(RecipeProjectError, match="must be absolute"):
        discover_recipe_project(recipe)


@pytest.mark.skipif(os.name != "posix", reason="symlinks are POSIX-flavored here")
def test_discover_symlink_escape_rejected(tmp_path: pathlib.Path) -> None:
    project = tmp_path / "proj"
    project.mkdir()
    _write_marker(project, 'format = "v1"\n')

    outside = tmp_path / "outside"
    outside.mkdir()
    real_recipe = outside / "x.star"
    real_recipe.write_text("")

    link = project / "x.star"
    link.symlink_to(real_recipe)

    with pytest.raises(RecipeProjectError, match="escapes its project root"):
        discover_recipe_project(link)


def test_safe_join_ok(tmp_path: pathlib.Path) -> None:
    (tmp_path / "lib").mkdir()
    target = tmp_path / "lib" / "docker.star"
    target.write_text("")
    assert safe_join(tmp_path, "lib/docker.star") == target.resolve()


def test_safe_join_rejects_traversal(tmp_path: pathlib.Path) -> None:
    with pytest.raises(RecipeProjectError, match="escapes recipe project root"):
        safe_join(tmp_path, "../evil")


def test_safe_join_rejects_absolute(tmp_path: pathlib.Path) -> None:
    with pytest.raises(RecipeProjectError, match="absolute paths are not allowed"):
        safe_join(tmp_path, "/etc/passwd")


def test_recipe_project_marker_path(tmp_path: pathlib.Path) -> None:
    _write_marker(tmp_path, 'format = "v1"\n')
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    rp = discover_recipe_project(recipe)
    assert isinstance(rp, RecipeProject)
    assert rp.marker_path == tmp_path.resolve() / MARKER_FILENAME


def test_discover_invalid_project_not_table(tmp_path: pathlib.Path) -> None:
    _write_marker(tmp_path, 'format = "v1"\nproject = 1\n')
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    with pytest.raises(RecipeProjectError, match=r"\[project\] must be a table"):
        discover_recipe_project(recipe)


def test_discover_invalid_project_name_empty(tmp_path: pathlib.Path) -> None:
    _write_marker(tmp_path, 'format = "v1"\n[project]\nname = ""\n')
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    with pytest.raises(RecipeProjectError, match="project.name"):
        discover_recipe_project(recipe)


def test_discover_invalid_project_name_non_string(tmp_path: pathlib.Path) -> None:
    _write_marker(tmp_path, 'format = "v1"\n[project]\nname = 123\n')
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    with pytest.raises(RecipeProjectError, match="project.name"):
        discover_recipe_project(recipe)


def test_discover_invalid_extra_artifact_roots_non_list(
    tmp_path: pathlib.Path,
) -> None:
    _write_marker(
        tmp_path,
        'format = "v1"\n[project]\nextra_artifact_roots = "scratch"\n',
    )
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    with pytest.raises(RecipeProjectError, match="extra_artifact_roots"):
        discover_recipe_project(recipe)


def test_discover_invalid_extra_artifact_roots_non_string_item(
    tmp_path: pathlib.Path,
) -> None:
    _write_marker(
        tmp_path,
        'format = "v1"\n[project]\nextra_artifact_roots = ["/ok", 1]\n',
    )
    recipe = tmp_path / "x.star"
    recipe.write_text("")
    with pytest.raises(RecipeProjectError, match="extra_artifact_roots"):
        discover_recipe_project(recipe)



from rich.console import Console, ConsoleOptions, RenderResult
from rich.markdown import CodeBlock, Heading, Markdown, MarkdownContext
from rich.syntax import Syntax
from rich.text import Text


class SlimHeading(Heading):
    def on_enter(self, context: MarkdownContext) -> None:
        try:
            # the heading level is indicated in the tag name in rich >= 13.2.0,
            # e.g. self.tag == 'h1', but directly stored in earlier versions
            # as self.level.
            #
            # see https://github.com/Textualize/rich/commit/a20c3d5468d02a55
            heading_level = int(self.tag[1:])  # type: ignore[attr-defined,unused-ignore]
        except AttributeError:
            heading_level = self.level  # type: ignore[attr-defined,unused-ignore]

        context.enter_style(self.style_name)
        self.text = Text("#" * heading_level + " ", context.current_style)

    def __rich_console__(
        self,
        console: Console,
        options: ConsoleOptions,
    ) -> RenderResult:
        yield self.text


# inspired by https://github.com/Textualize/rich/issues/3154
class NonWrappingCodeBlock(CodeBlock):
    def __rich_console__(
        self,
        console: Console,
        options: ConsoleOptions,
    ) -> RenderResult:
        # re-enable non-wrapping options locally for code blocks
        render_options = options.update(no_wrap=True, overflow="ignore")

        code = str(self.text).rstrip()
        syntax = Syntax(
            code,
            self.lexer_name,
            theme=self.theme,
            word_wrap=False,
            # not supported in rich <= 12.4.0 (Textualize/rich#2247) but fortunately
            # zero padding is the default anyway
            # padding=0,
        )
        return syntax.highlight(code).__rich_console__(console, render_options)


class RuyiStyledMarkdown(Markdown):
    elements = Markdown.elements
    elements["fence"] = NonWrappingCodeBlock
    elements["heading_open"] = SlimHeading

    # rich < 13.2.0
    # see https://github.com/Textualize/rich/commit/745bd99e416c2806
    # it doesn't hurt to just unconditionally add them like below
    elements["code"] = NonWrappingCodeBlock
    elements["code_block"] = NonWrappingCodeBlock
    elements["heading"] = SlimHeading

    def __rich_console__(
        self,
        console: Console,
        options: ConsoleOptions,
    ) -> RenderResult:
        # we have to undo the ruyi-global console's non-wrapping setting
        # for proper CLI rendering of long lines
        render_options = options.update(no_wrap=False, overflow="fold")
        return super().__rich_console__(console, render_options)



import pathlib
from typing import Iterable, NotRequired, Sequence, TYPE_CHECKING

from .check import (
    CheckDiagnostic,
    CheckName,
    CheckSet,
    CheckSeverity,
    CheckUsageError,
    DEFAULT_CHECKS,
    CHECK_FORMAT,
    CHECK_PARSE,
    check_manifest_file,
    check_repo,
)
from .list_filter import ListFilter, ListFilterOp, ListFilterOpKind
from ..utils.porcelain import PorcelainEntity, PorcelainOutput

if TYPE_CHECKING:
    from ..config import GlobalConfig


class CheckDiagnosticPorcelain(PorcelainEntity, total=False):
    severity: str
    code: str
    check: str
    path: str
    message: str
    line: NotRequired[int]
    column: NotRequired[int]
    hint: NotRequired[str]


def normalize_checks(raw_checks: Iterable[str] | None) -> CheckSet:
    if raw_checks is None:
        return DEFAULT_CHECKS

    checks: set[CheckName] = set()
    for raw_check in raw_checks:
        match raw_check:
            case "format":
                checks.add(CHECK_FORMAT)
            case "parse":
                checks.add(CHECK_PARSE)
            case _:
                raise CheckUsageError(f"unsupported check: {raw_check}")
    return frozenset(checks) if checks else DEFAULT_CHECKS


def parse_package_selector_args(tokens: Sequence[str]) -> ListFilter:
    if not tokens:
        raise CheckUsageError("--only-packages requires at least one package selector")

    result = ListFilter()
    idx = 0
    unary_options = {
        "--category-contains": ListFilterOpKind.CATEGORY_CONTAINS,
        "--category-is": ListFilterOpKind.CATEGORY_IS,
        "--name-contains": ListFilterOpKind.NAME_CONTAINS,
    }
    unsupported_options = {"--is-installed", "--related-to-entity"}

    while idx < len(tokens):
        token = tokens[idx]
        option, sep, inline_value = token.partition("=")

        if option == "--all":
            if sep:
                raise CheckUsageError("--all does not take an argument")
            result.append(ListFilterOp(ListFilterOpKind.ALL, ""))
            idx += 1
            continue

        if option in unsupported_options:
            raise CheckUsageError(
                f"package selector {option} is not supported by admin check"
            )

        op_kind = unary_options.get(option)
        if op_kind is None:
            raise CheckUsageError(f"unsupported package selector: {token}")

        if sep:
            value = inline_value
        else:
            idx += 1
            if idx >= len(tokens):
                raise CheckUsageError(f"package selector {option} requires an argument")
            value = tokens[idx]
        result.append(ListFilterOp(op_kind, value))
        idx += 1

    return result


def format_check_diagnostic(diagnostic: CheckDiagnostic) -> str:
    location = str(diagnostic.path)
    if diagnostic.line is not None:
        location += f":{diagnostic.line}"
        if diagnostic.column is not None:
            location += f":{diagnostic.column}"

    result = (
        f"{location}: {diagnostic.severity.value} {diagnostic.code}: "
        f"{diagnostic.message}"
    )
    if diagnostic.hint is not None:
        result += f"\nhint: {diagnostic.hint}"
    return result


def count_diagnostics(
    diagnostics: Iterable[CheckDiagnostic],
) -> tuple[int, int]:
    errors = 0
    warnings = 0
    for diagnostic in diagnostics:
        match diagnostic.severity:
            case CheckSeverity.ERROR:
                errors += 1
            case CheckSeverity.WARNING:
                warnings += 1
    return errors, warnings


def format_check_summary(diagnostics: Iterable[CheckDiagnostic]) -> str:
    errors, warnings = count_diagnostics(diagnostics)
    return f"{errors} error(s), {warnings} warning(s)"


def has_error_diagnostic(diagnostics: Iterable[CheckDiagnostic]) -> bool:
    return any(diagnostic.severity == CheckSeverity.ERROR for diagnostic in diagnostics)


def do_admin_check(
    cfg: "GlobalConfig",
    *,
    files: Sequence[str] | None,
    repo: str | None,
    checks: Iterable[str] | None,
    only_packages: Sequence[str] | None,
) -> int:
    check_set = normalize_checks(checks)
    package_selector = (
        parse_package_selector_args(only_packages)
        if only_packages is not None
        else None
    )

    diagnostics: list[CheckDiagnostic] = []
    if repo is not None:
        diagnostics.extend(
            check_repo(
                pathlib.Path(repo),
                checks=check_set,
                package_selector=package_selector,
            )
        )
    else:
        if not files:
            raise CheckUsageError("either --file or --repo must be specified")
        for file in files:
            diagnostics.extend(check_manifest_file(pathlib.Path(file), check_set))

    if cfg.is_porcelain:
        with PorcelainOutput() as po:
            for diagnostic in diagnostics:
                po.emit(diagnostic.to_porcelain())
    else:
        for diagnostic in diagnostics:
            cfg.logger.stdout(format_check_diagnostic(diagnostic))
        cfg.logger.stdout(format_check_summary(diagnostics))

    return 1 if has_error_diagnostic(diagnostics) else 0



import argparse
import pathlib
from typing import TYPE_CHECKING

from ..cli.cmd import RootCommand
from ..i18n import _
from .cli_completion import package_completer_builder
from .host import get_native_host

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class ExtractCommand(
    RootCommand,
    cmd="extract",
    help=_("Fetch package(s) then extract to current directory"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        a = p.add_argument(
            "atom",
            type=str,
            nargs="+",
            help=_("Specifier (atom) of the package(s) to extract"),
        )
        if gc.is_cli_autocomplete:
            a.completer = package_completer_builder(gc)

        p.add_argument(
            "-d",
            "--dest-dir",
            type=str,
            metavar="DESTDIR",
            default=".",
            help=_("Destination directory to extract to (default: current directory)"),
        )
        p.add_argument(
            "--extract-without-subdir",
            action="store_true",
            help=_(
                "Extract files directly into DESTDIR instead of package-named subdirectories"
            ),
        )
        p.add_argument(
            "-f",
            "--fetch-only",
            action="store_true",
            help=_("Fetch distribution files only without installing"),
        )
        p.add_argument(
            "--host",
            type=str,
            default=get_native_host(),
            help=_("Override the host architecture (normally not needed)"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .host import canonicalize_host_str
        from .install import do_extract_atoms

        atom_strs: set[str] = set(args.atom)
        dest_dir_arg: str = args.dest_dir
        extract_without_subdir: bool = args.extract_without_subdir
        host: str = args.host
        fetch_only: bool = args.fetch_only

        dest_dir = None if dest_dir_arg == "." else pathlib.Path(dest_dir_arg)

        return do_extract_atoms(
            cfg,
            cfg.repo,
            atom_strs,
            canonicalized_host=canonicalize_host_str(host),
            dest_dir=dest_dir,
            extract_without_subdir=extract_without_subdir,
            fetch_only=fetch_only,
        )


class InstallCommand(
    RootCommand,
    cmd="install",
    aliases=["i"],
    help=_("Install package from configured repository"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        a = p.add_argument(
            "atom",
            type=str,
            nargs="+",
            help=_("Specifier (atom) of the package to install"),
        )
        if gc.is_cli_autocomplete:
            a.completer = package_completer_builder(gc)

        p.add_argument(
            "-f",
            "--fetch-only",
            action="store_true",
            help=_("Fetch distribution files only without installing"),
        )
        p.add_argument(
            "--host",
            type=str,
            default=get_native_host(),
            help=_("Override the host architecture (normally not needed)"),
        )
        p.add_argument(
            "--reinstall",
            action="store_true",
            help=_("Force re-installation of already installed packages"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .host import canonicalize_host_str
        from .install import do_install_atoms

        host: str = args.host
        atom_strs: set[str] = set(args.atom)
        fetch_only: bool = args.fetch_only
        reinstall: bool = args.reinstall

        return do_install_atoms(
            cfg,
            cfg.repo,
            atom_strs,
            canonicalized_host=canonicalize_host_str(host),
            fetch_only=fetch_only,
            reinstall=reinstall,
        )


class UninstallCommand(
    RootCommand,
    cmd="uninstall",
    aliases=["remove", "rm"],
    help=_("Uninstall installed packages"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: argparse.ArgumentParser) -> None:
        p.add_argument(
            "atom",
            type=str,
            nargs="+",
            help=_("Specifier (atom) of the package to uninstall"),
        )
        p.add_argument(
            "--host",
            type=str,
            default=get_native_host(),
            help=_("Override the host architecture (normally not needed)"),
        )
        p.add_argument(
            "-y",
            "--yes",
            action="store_true",
            dest="assume_yes",
            help=_("Assume yes to all prompts"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .host import canonicalize_host_str
        from .install import do_uninstall_atoms

        host: str = args.host
        atom_strs: set[str] = set(args.atom)
        assume_yes: bool = args.assume_yes

        return do_uninstall_atoms(
            cfg,
            cfg.repo,
            atom_strs,
            canonicalized_host=canonicalize_host_str(host),
            assume_yes=assume_yes,
        )



#!/usr/bin/env python3

import os
import pathlib
import shutil
import sys
import tomllib


def main(argv: list[str]) -> int:
    if len(argv) != 2:
        print(f"usage: {argv[0]} <release-artifacts-dir>", file=sys.stderr)
        return 1

    workdir = pathlib.Path(argv[1]).resolve()

    project_root = (pathlib.Path(os.path.dirname(__file__)) / "..").resolve()
    with open(project_root / "pyproject.toml", "rb") as fp:
        pyproject = tomllib.load(fp)

    try:
        version = pyproject["project"]["version"]
    except KeyError:
        # In case the packaging environment has Poetry 1.x metadata switched
        # in
        version = pyproject["tool"]["poetry"]["version"]

    # layout of release-artifacts-dir just after the download-artifacts@v4
    # action:
    #
    # release-artifacts-dir
    # ├── ruyi-XXXXXXXX.tar.gz
    # │   └── ruyi-XXXXXXXX.tar.gz
    # ├── ruyi.amd64
    # │   └── ruyi
    # ├── ruyi.arm64
    # │   └── ruyi
    # ├── ruyi.riscv64
    # │   └── ruyi
    # └── ruyi.windows-amd64.exe
    #     └── ruyi.exe
    #
    # we want to organize it into the following layout:
    #
    # release-artifacts-dir
    # ├── ruyi-XXXXXXXX.tar.gz
    # ├── ruyi-<semver>.amd64
    # ├── ruyi-<semver>.arm64
    # └── ruyi-<semver>.riscv64
    #
    # i.e. with the non-Linux build removed, with the directory structure
    # flattened, and with the semver attached.

    os.chdir(workdir)

    # for now, hardcode the exact artifacts we want
    included_arches = ("amd64", "arm64", "riscv64")
    wanted_names = {f"ruyi.{arch}" for arch in included_arches}
    names = os.listdir(".")
    for name in names:
        if name.endswith(".tar.gz"):
            src_path = os.path.join(name, name)
            tmp_path = f"{name}.new"
            print(f"moving tarball {src_path} outside")
            os.rename(src_path, tmp_path)
            os.rmdir(name)
            os.rename(tmp_path, name)
            continue

        if not name.startswith("ruyi"):
            print(f"ignoring {name}")
            continue

        if name not in wanted_names:
            print(f"removing unwanted {name}")
            shutil.rmtree(name)
            continue

        # assume name is ruyi.{arch}
        arch = name.rsplit(".", 1)[1]
        src_name = os.path.join(name, "ruyi")
        dest_name = f"ruyi-{version}.{arch}"
        print(f"moving {src_name} to {dest_name}")
        os.rename(src_name, dest_name)
        os.chmod(dest_name, 0o755)
        os.rmdir(name)

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3

import os
import tomllib
from typing import cast


def main() -> None:
    v = get_semver()
    set_release_mirror_url_for_gha(v)


def get_semver() -> str:
    # assume CWD is project root, which is guaranteed to be the case (see
    # end of file)
    with open("pyproject.toml", "rb") as fp:
        pyproject = tomllib.load(fp)

    return cast(str, pyproject["project"]["version"])


def set_release_mirror_url_for_gha(version: str) -> None:
    release_url_base = "https://mirror.iscas.ac.cn/ruyisdk/ruyi/releases/"
    testing_url_base = "https://mirror.iscas.ac.cn/ruyisdk/ruyi/testing/"

    # Do not depend on external libraries so this can work in plain GHA
    # environment without any venv setup. See the SemVer spec -- as long as
    # we don't have build tags containing "-" we should be fine, which is
    # exactly the case.
    #
    # sv = Version.parse(version)
    # is_prerelease = sv.prerelease
    is_prerelease = "-" in version

    url_base = testing_url_base if is_prerelease else release_url_base
    url = f"{url_base}{version}/"
    set_gha_output("release_mirror_url", url)


def set_gha_output(k: str, v: str) -> None:
    if "\n" in v:
        raise ValueError("this helper is only for small one-line outputs")

    # only do this when the GitHub Actions output file is available
    # https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#setting-an-output-parameter
    outfile = os.environ.get("GITHUB_OUTPUT", "")
    if not outfile:
        return

    with open(outfile, "a", encoding="utf-8") as fp:
        fp.write(f"{k}={v}\n")


if __name__ == "__main__":
    # cd to project root
    os.chdir(os.path.join(os.path.dirname(__file__), ".."))
    main()



from itertools import chain

from ..config import GlobalConfig
from ..i18n import _
from ..log import RuyiLogger
from ..utils.porcelain import PorcelainOutput
from .augmented_pkg import AugmentedPkg
from .list_filter import ListFilter
from .pkg_manifest import BoundPackageManifest


def do_list(
    cfg: GlobalConfig,
    filters: ListFilter,
    verbose: bool,
) -> int:
    logger = cfg.logger

    if not filters:
        if cfg.is_porcelain:
            # we don't want to print message for humans in case of porcelain
            # mode, but we don't want to retain the old behavior of listing
            # all packages either
            return 1

        logger.F(_("no filter specified for list operation"))
        logger.I(
            _(
                "for the old behavior of listing all packages, try [yellow]ruyi list --all[/]"
            )
        )
        return 1

    augmented_pkgs = list(AugmentedPkg.yield_from_repo(cfg, cfg.repo, filters))
    multi_repo = len(cfg.repo_entries) > 1

    if cfg.is_porcelain:
        return _do_list_porcelain(augmented_pkgs)

    if not verbose:
        return _do_list_non_verbose(logger, augmented_pkgs, multi_repo)

    for i, ver in enumerate(chain(*(ap.versions for ap in augmented_pkgs))):
        if i > 0:
            logger.stdout("\n")

        _print_pkg_detail(logger, ver.pm, cfg.lang_code, multi_repo)

    return 0


def _do_list_non_verbose(
    logger: RuyiLogger,
    augmented_pkgs: list[AugmentedPkg],
    multi_repo: bool = False,
) -> int:
    logger.stdout(_("List of available packages:\n"))

    for ap in augmented_pkgs:
        logger.stdout(f"* [bold green]{ap.category}/{ap.name}[/]")
        for ver in ap.versions:
            if ver.remarks:
                comments_str = (
                    f" ({', '.join(r.as_rich_markup() for r in ver.remarks)})"
                )
            else:
                comments_str = ""
            slug_str = f" slug: [yellow]{ver.pm.slug}[/]" if ver.pm.slug else ""
            repo_str = f" [dim]\\[{ver.pm.repo_id}][/]" if multi_repo else ""
            logger.stdout(
                f"  - [blue]{ver.pm.semver}[/]{comments_str}{slug_str}{repo_str}"
            )

    return 0


def _do_list_porcelain(augmented_pkgs: list[AugmentedPkg]) -> int:
    with PorcelainOutput() as po:
        for ap in augmented_pkgs:
            po.emit(ap.to_porcelain())

    return 0


def _print_pkg_detail(
    logger: RuyiLogger,
    pm: BoundPackageManifest,
    lang_code: str,
    multi_repo: bool = False,
) -> None:
    repo_tag = f" [dim]\\[{pm.repo_id}][/]" if multi_repo else ""
    logger.stdout(
        f"[bold]## [green]{pm.category}/{pm.name}[/] [blue]{pm.ver}[/]{repo_tag}[/]\n"
    )

    if pm.slug is not None:
        logger.stdout(_("* Slug: [yellow]{slug}[/]").format(slug=pm.slug))
    else:
        logger.stdout(_("* Slug: (none)"))
    logger.stdout(_("* Package kind: {kind}").format(kind=sorted(pm.kind)))
    logger.stdout(_("* Vendor: {vendor}").format(vendor=pm.vendor_name))
    if upstream_ver := pm.upstream_version:
        logger.stdout(
            _("* Upstream version number: {version}").format(version=upstream_ver)
        )
    else:
        logger.stdout(_("* Upstream version number: (undeclared)"))
    logger.stdout("")

    sv = pm.service_level
    if sv.has_known_issues:
        logger.stdout(_("\nPackage has known issue(s):\n"))
        for x in sv.render_known_issues(pm.repo.messages, lang_code):
            logger.stdout(x, end="\n\n")

    df = pm.distfiles
    logger.stdout(_("Package declares {count} distfile(s):\n").format(count=len(df)))
    for dd in df.values():
        logger.stdout(f"* [green]{dd.name}[/]")
        logger.stdout(_("    - Size: [yellow]{size}[/] bytes").format(size=dd.size))
        for kind, csum in dd.checksums.items():
            logger.stdout(f"    - {kind.upper()}: [yellow]{csum}[/]")

    if bm := pm.binary_metadata:
        logger.stdout(_("\n### Binary artifacts\n"))
        for host, data in bm.data.items():
            logger.stdout(_("* Host [green]{host}[/]:").format(host=host))
            logger.stdout(
                _("    - Distfiles: {distfiles}").format(distfiles=data["distfiles"])
            )
            if cmds := data.get("commands"):
                logger.stdout(_("    - Available command(s):"))
                for k in sorted(cmds.keys()):
                    logger.stdout(f"        - [green]{k}[/]")

    if tm := pm.toolchain_metadata:
        logger.stdout(_("\n### Toolchain metadata\n"))
        logger.stdout(_("* Target: [bold green]{target}[/]").format(target=tm.target))
        logger.stdout(_("* Quirks: {quirks}").format(quirks=tm.quirks))
        logger.stdout(_("* Components:"))
        for tc in tm.components:
            logger.stdout(f'    - {tc["name"]} [bold green]{tc["version"]}[/]')



import io
import types
from typing import BinaryIO

HEADER_BSD: int
HEADER_GNU: int
HEADER_GNU_TABLE: int
HEADER_GNU_SYMBOLS: int
HEADER_NORMAL: int
HEADER_TYPES: dict[int, str]
GLOBAL_HEADER_LEN: int
HEADER_LEN: int

class ArchiveFormatError(Exception): ...
class ArchiveAccessError(IOError): ...

class ArchiveFileHeader:
    type: int
    size: int
    timestamp: int
    uid: int | None
    gid: int | None
    mode: int
    offset: int
    name: bytes
    file_offset: int | None
    proxy_name: bytes
    def __init__(self, header: bytes, offset: int) -> None: ...

class ArchiveFileData(io.IOBase):
    header: ArchiveFileHeader
    arobj: Archive
    last_offset: int
    def __init__(self, ar_obj: Archive, header: ArchiveFileHeader) -> None: ...
    def read(self, size: int | None = None) -> bytes: ...
    def tell(self) -> int: ...
    def seek(self, offset: int, whence: int = 0) -> int: ...
    def seekable(self) -> bool: ...
    def __enter__(self) -> ArchiveFileData: ...
    def __exit__(
        self,
        _exc_type: type[BaseException] | None,
        _exc_value: BaseException | None,
        _traceback: types.TracebackType | None,
    ) -> None: ...

class ArchiveFileDataThin(ArchiveFileData):
    file_path: str
    def __init__(self, ar_obj: Archive, header: ArchiveFileHeader) -> None: ...
    def read(self, size: int | None = None) -> bytes: ...

class Archive:
    headers: list[ArchiveFileHeader]
    file: BinaryIO
    position: int
    reached_eof: bool
    file_data_class: type[ArchiveFileData] | type[ArchiveFileDataThin]
    next_header_offset: int
    gnu_table: dict[int, bytes]
    archived_files: dict[bytes, ArchiveFileData]
    def __init__(
        self, filename: str | None = None, fileobj: BinaryIO | None = None
    ) -> None: ...
    def read_next_header(self) -> ArchiveFileHeader | None: ...
    def __next__(self) -> ArchiveFileData: ...
    next = __next__
    def __iter__(self) -> Archive: ...
    def read_all_headers(self) -> None: ...
    def close(self) -> None: ...
    def namelist(self) -> list[bytes]: ...
    def infolist(self) -> list[ArchiveFileHeader]: ...
    def open(self, name: bytes | ArchiveFileHeader) -> ArchiveFileData: ...
    def __enter__(self) -> Archive: ...
    def __exit__(
        self,
        _exc_type: type[BaseException] | None,
        _exc_value: BaseException | None,
        _traceback: types.TracebackType | None,
    ) -> None: ...



import argparse
import os
import pathlib
import shutil

import pytest

from ruyi.cli.completion import ArgumentParser
from ruyi.mux.venv.maker import (
    SysrootProvisionMode,
    VenvProvisionError,
    provision_sysroot,
)
from ruyi.mux.venv.venv_cli import VenvCommand
from tests.fixtures import IntegrationTestHarness


def make_parser(ruyi_cli_runner: IntegrationTestHarness) -> ArgumentParser:
    ctx = ruyi_cli_runner.make_command_context("venv")
    p = ruyi_cli_runner.make_parser()
    VenvCommand.configure_args(ctx.gc, p)
    return p


def test_parse_new_sysroot_source_flags(
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    p = make_parser(ruyi_cli_runner)

    args = p.parse_args(
        [
            "default",
            "./venv",
            "--copy-sysroot-from-dir",
            "/tmp/sysroot",
        ]
    )
    assert args.copy_sysroot_from_dir == "/tmp/sysroot"
    assert args.symlink_sysroot_from_dir is None
    assert args.copy_sysroot_from_pkg is None
    assert args.project_sysroot_from_rootfs is None

    args = p.parse_args(
        [
            "default",
            "./venv",
            "--symlink-sysroot-from-dir",
            "/tmp/sysroot",
        ]
    )
    assert args.symlink_sysroot_from_dir == "/tmp/sysroot"

    args = p.parse_args(
        [
            "default",
            "./venv",
            "--project-sysroot-from-rootfs",
            "/tmp/rootfs",
        ]
    )
    assert args.project_sysroot_from_rootfs == "/tmp/rootfs"


def test_old_sysroot_from_alias_still_parses(
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    p = make_parser(ruyi_cli_runner)

    args = p.parse_args(
        [
            "default",
            "./venv",
            "--sysroot-from",
            "gnu-plct",
        ]
    )
    assert args.copy_sysroot_from_pkg == "gnu-plct"


def test_help_documents_sysroot_provisioning(
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    p = make_parser(ruyi_cli_runner)

    help_text = p.format_help()

    assert "Sysroot provisioning:" in help_text
    assert "--copy-sysroot-from-dir only for a complete sysroot directory" in help_text
    assert (
        "--project-sysroot-from-rootfs for distro rootfs or chroot trees" in help_text
    )
    assert (
        "Ruyi never elevates privileges when creating virtual environments" in help_text
    )


def test_sysroot_source_options_are_mutually_exclusive_at_main(
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    ctx = ruyi_cli_runner.make_command_context("venv")
    args = argparse.Namespace(
        profile="default",
        dest="./venv",
        with_sysroot=True,
        name=None,
        toolchain=["gnu-plct"],
        emulator=None,
        copy_sysroot_from_pkg="foo",
        copy_sysroot_from_dir="/tmp/sysroot",
        symlink_sysroot_from_dir=None,
        project_sysroot_from_rootfs=None,
        extra_commands_from=None,
    )

    rc = VenvCommand.main(ctx.gc, args)
    assert rc == 1
    assert ctx.fatal_messages == [
        "at most one of --copy-sysroot-from-pkg, --copy-sysroot-from-dir, --symlink-sysroot-from-dir, and --project-sysroot-from-rootfs may be specified"
    ]


def test_without_sysroot_conflicts_with_explicit_source(
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    ctx = ruyi_cli_runner.make_command_context("venv")
    args = argparse.Namespace(
        profile="default",
        dest="./venv",
        with_sysroot=False,
        name=None,
        toolchain=["gnu-plct"],
        emulator=None,
        copy_sysroot_from_pkg=None,
        copy_sysroot_from_dir="/tmp/sysroot",
        symlink_sysroot_from_dir=None,
        project_sysroot_from_rootfs=None,
        extra_commands_from=None,
    )

    rc = VenvCommand.main(ctx.gc, args)
    assert rc == 1
    assert ctx.fatal_messages == [
        "--without-sysroot cannot be combined with a sysroot source option"
    ]


def test_copy_sysroot_failure_reports_clean_diagnostic(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    ctx = ruyi_cli_runner.make_command_context("venv")
    src = tmp_path / "sysroot"
    dest = tmp_path / "venv" / "sysroot.riscv64-test-linux-gnu"
    src.mkdir()

    def fake_copytree(*args: object, **kwargs: object) -> None:
        raise shutil.Error(
            [
                (
                    str(src / "etc" / "shadow"),
                    str(dest / "etc" / "shadow"),
                    "permission denied",
                ),
            ]
        )

    monkeypatch.setattr(shutil, "copytree", fake_copytree)

    with pytest.raises(VenvProvisionError):
        provision_sysroot(
            ctx.gc.logger,
            src,
            dest,
            SysrootProvisionMode.COPY_TREE,
            "riscv64-test-linux-gnu",
        )

    assert ctx.fatal_messages == [
        f"cannot copy sysroot from {src}: one entry could not be copied"
    ]
    assert (
        "Ruyi does not elevate privileges when creating virtual environments"
        in ctx.stderr.getvalue()
    )


def test_project_sysroot_from_rootfs_copies_common_roots(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    ctx = ruyi_cli_runner.make_command_context("venv")
    src = tmp_path / "rootfs"
    dest = tmp_path / "venv" / "sysroot.riscv64-test-linux-gnu"

    (src / "usr" / "include").mkdir(parents=True)
    (src / "usr" / "include" / "game.h").write_text("#pragma once\n", encoding="utf-8")
    (src / "usr" / "lib").mkdir(parents=True)
    (src / "usr" / "lib" / "libgame.so").write_bytes(b"fake so")
    (src / "usr" / "lib" / "ld-linux-riscv64-lp64d.so.1").write_bytes(b"fake ld")
    (src / "lib64").mkdir()
    os.symlink(
        "/usr/lib/ld-linux-riscv64-lp64d.so.1",
        src / "lib64" / "ld-linux-riscv64-lp64d.so.1",
    )
    (src / "etc").mkdir()
    (src / "etc" / "shadow").write_text("should not be copied\n", encoding="utf-8")

    provision_sysroot(
        ctx.gc.logger,
        src,
        dest,
        SysrootProvisionMode.PROJECT_ROOTFS,
        "riscv64-test-linux-gnu",
    )

    assert (dest / "usr" / "include" / "game.h").read_text(
        encoding="utf-8"
    ) == "#pragma once\n"
    assert (dest / "usr" / "lib" / "libgame.so").read_bytes() == b"fake so"
    assert not (dest / "etc" / "shadow").exists()
    assert os.readlink(dest / "lib64" / "ld-linux-riscv64-lp64d.so.1") == (
        "../usr/lib/ld-linux-riscv64-lp64d.so.1"
    )


def test_project_sysroot_from_rootfs_skips_unsupported_entries(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    ctx = ruyi_cli_runner.make_command_context("venv")
    src = tmp_path / "rootfs"
    dest = tmp_path / "venv" / "sysroot.riscv64-test-linux-gnu"
    (src / "usr" / "lib").mkdir(parents=True)
    (src / "usr" / "lib" / "libgame.so").write_bytes(b"fake so")
    os.mkfifo(src / "usr" / "lib" / "unsupported-fifo")

    provision_sysroot(
        ctx.gc.logger,
        src,
        dest,
        SysrootProvisionMode.PROJECT_ROOTFS,
        "riscv64-test-linux-gnu",
    )

    assert (dest / "usr" / "lib" / "libgame.so").read_bytes() == b"fake so"
    assert not (dest / "usr" / "lib" / "unsupported-fifo").exists()
    assert "some unreadable or unsupported files were skipped" in ctx.stderr.getvalue()


def test_project_sysroot_from_rootfs_fails_without_supported_roots(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    ctx = ruyi_cli_runner.make_command_context("venv")
    src = tmp_path / "rootfs"
    dest = tmp_path / "venv" / "sysroot.riscv64-test-linux-gnu"
    (src / "etc").mkdir(parents=True)

    with pytest.raises(VenvProvisionError):
        provision_sysroot(
            ctx.gc.logger,
            src,
            dest,
            SysrootProvisionMode.PROJECT_ROOTFS,
            "riscv64-test-linux-gnu",
        )

    assert ctx.fatal_messages == [
        f"cannot project sysroot from {src}: no supported sysroot directories were found"
    ]



from typing import Any, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from ..cli.completer import DynamicCompleter
    from ..config import GlobalConfig


def repo_id_completer_builder(
    cfg: "GlobalConfig",
) -> "DynamicCompleter":
    repo_ids = [entry.id for entry in cfg.repo_entries]

    def f(prefix: str, parsed_args: object, **kwargs: Any) -> list[str]:
        return [rid for rid in repo_ids if rid.startswith(prefix)]

    return f


def package_completer_builder(
    cfg: "GlobalConfig",
    filters: list[Callable[[str], bool]] | None = None,
) -> "DynamicCompleter":
    pkg_names: list[str] | None = None

    def f(prefix: str, parsed_args: object, **kwargs: Any) -> list[str]:
        nonlocal pkg_names

        if pkg_names is None:
            # Lazy import to avoid circular dependency, and lazy repo access so
            # parser construction for unrelated completions does not sync repos.
            from ..ruyipkg.augmented_pkg import (
                AugmentedPkg,
            )  # pylint: disable=import-outside-toplevel
            from ..ruyipkg.list_filter import (
                ListFilter,
            )  # pylint: disable=import-outside-toplevel

            pkg_names = []
            for pkg in AugmentedPkg.yield_from_repo(cfg, cfg.repo, ListFilter()):
                if pkg.name is None:
                    continue
                if filters is not None and not all(fn(pkg.name) for fn in filters):
                    continue
                pkg_names.append(pkg.name)

        return [name for name in pkg_names if name.startswith(prefix)]

    return f



import io
import hashlib

import pytest

from ruyi.ruyipkg.checksum import get_hash_instance, Checksummer


def test_get_hash_instance_supported() -> None:
    # these should not raise any exception
    get_hash_instance("sha256")
    get_hash_instance("sha512")


def test_get_hash_instance_unsupported() -> None:
    with pytest.raises(ValueError, match="checksum algorithm md5 not supported"):
        get_hash_instance("md5")


def test_checksummer_compute() -> None:
    file_content = b"test content"
    expected_sha256 = hashlib.sha256(file_content).hexdigest()
    expected_sha512 = hashlib.sha512(file_content).hexdigest()

    file = io.BytesIO(file_content)
    checksums = {"sha256": expected_sha256, "sha512": expected_sha512}
    checksummer = Checksummer(file, checksums)

    computed_checksums = checksummer.compute()
    assert computed_checksums["sha256"] == expected_sha256


def test_checksummer_check() -> None:
    file_content = b"test content"
    expected_sha256 = hashlib.sha256(file_content).hexdigest()
    expected_sha512 = hashlib.sha512(file_content).hexdigest()

    file = io.BytesIO(file_content)
    checksums = {"sha256": expected_sha256, "sha512": expected_sha512}
    checksummer = Checksummer(file, checksums)

    # This should not raise any exception
    checksummer.check()

    # Modify the file content to cause a checksum mismatch
    file = io.BytesIO(b"modified content")
    checksummer = Checksummer(file, checksums)

    with pytest.raises(ValueError, match="wrong sha256 checksum"):
        checksummer.check()



import argparse
import pathlib
from typing import TYPE_CHECKING

from ..cli.cmd import RootCommand
from ..i18n import _

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class RepoCommand(
    RootCommand,
    cmd="repo",
    has_subcommands=True,
    help=_("Manage configured package repositories"),
):
    @classmethod
    def configure_args(
        cls,
        gc: "GlobalConfig",
        p: "ArgumentParser",
    ) -> None:
        pass


class RepoListCommand(
    RepoCommand,
    cmd="list",
    help=_("List configured package repositories"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .repo import DEFAULT_REPO_ID
        from ..utils.porcelain import PorcelainOutput

        entries = cfg.repo_entries
        logger = cfg.logger

        if cfg.is_porcelain:
            with PorcelainOutput() as po:
                for entry in entries:
                    po.emit(entry.to_porcelain())
            return 0

        for entry in sorted(entries, key=lambda e: -e.priority):
            active_marker = "*" if entry.active else " "
            default_marker = " (default)" if entry.id == DEFAULT_REPO_ID else ""
            system_marker = " (system)" if entry.is_system else ""

            source = entry.remote or ""
            if entry.local_path:
                source = (
                    entry.local_path
                    if not source
                    else f"{source} (local: {entry.local_path})"
                )

            logger.stdout(
                f"  {active_marker} [bold]{entry.id}[/]{default_marker}{system_marker}  "
                f"priority={entry.priority}  {source}"
            )

        return 0


class RepoAddCommand(
    RepoCommand,
    cmd="add",
    help=_("Add a package repository"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument("id", type=str, help=_("unique repository identifier"))
        p.add_argument(
            "url", type=str, nargs="?", default=None, help=_("git remote URL")
        )
        p.add_argument(
            "--branch", type=str, default=None, help=_("git branch to track")
        )
        p.add_argument(
            "--priority",
            type=int,
            default=0,
            help=_("priority (higher = overrides lower)"),
        )
        p.add_argument(
            "--local",
            type=str,
            default=None,
            help=_("local path to use instead of or alongside remote"),
        )
        p.add_argument(
            "--name", type=str, default=None, help=_("human-readable name for the repo")
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from ..config.editor import ConfigEditor
        from ..config.schema import (
            KEY_REPOS_ACTIVE,
            KEY_REPOS_BRANCH,
            KEY_REPOS_ID,
            KEY_REPOS_LOCAL,
            KEY_REPOS_NAME,
            KEY_REPOS_PRIORITY,
            KEY_REPOS_REMOTE,
        )
        from .repo import DEFAULT_REPO_ID, REPO_ID_PATTERN

        logger = cfg.logger
        repo_id: str = args.id
        url: str | None = args.url
        local: str | None = args.local

        if not REPO_ID_PATTERN.match(repo_id):
            logger.F(_("invalid repo id '{id}'").format(id=repo_id))
            return 1

        if repo_id == DEFAULT_REPO_ID:
            logger.F(
                _(
                    "'{id}' is reserved; use [repo] config to configure the default repository"
                ).format(
                    id=DEFAULT_REPO_ID,
                )
            )
            return 1

        if not url and not local:
            logger.F(_("at least one of URL or --local must be provided"))
            return 1

        if local and not pathlib.Path(local).is_absolute():
            logger.F(_("local path '{path}' must be absolute").format(path=local))
            return 1

        # Check for conflict with existing entries.
        for entry in cfg.repo_entries:
            if entry.id == repo_id:
                logger.F(_("a repo with id '{id}' already exists").format(id=repo_id))
                return 1

        entry_data: dict[str, object] = {KEY_REPOS_ID: repo_id}
        if args.name:
            entry_data[KEY_REPOS_NAME] = args.name
        if url:
            entry_data[KEY_REPOS_REMOTE] = url
        if args.branch:
            entry_data[KEY_REPOS_BRANCH] = args.branch
        if local:
            entry_data[KEY_REPOS_LOCAL] = local
        if args.priority != 0:
            entry_data[KEY_REPOS_PRIORITY] = args.priority
        entry_data[KEY_REPOS_ACTIVE] = True

        with ConfigEditor.work_on_user_local_config(cfg) as editor:
            editor.add_repos_entry(entry_data)
            editor.stage()

        logger.I(_("repo '{id}' added; run 'ruyi update' to sync").format(id=repo_id))
        return 0


class RepoRemoveCommand(
    RepoCommand,
    cmd="remove",
    help=_("Remove a package repository"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        a = p.add_argument("id", type=str, help=_("repository identifier to remove"))
        if gc.is_cli_autocomplete:
            from .cli_completion import repo_id_completer_builder

            a.completer = repo_id_completer_builder(gc)
        p.add_argument(
            "--purge",
            action="store_true",
            default=False,
            help=_("also remove cached repo data from disk"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        import shutil

        from ..config.editor import ConfigEditor
        from .repo import DEFAULT_REPO_ID

        logger = cfg.logger
        repo_id: str = args.id

        if repo_id == DEFAULT_REPO_ID:
            logger.F(
                _(
                    "cannot remove the default repo '{id}'; use 'repo disable' instead"
                ).format(
                    id=DEFAULT_REPO_ID,
                )
            )
            return 1

        # Check if entry is system-provided
        for entry in cfg.repo_entries:
            if entry.id == repo_id and entry.is_system:
                logger.F(
                    _(
                        "cannot remove system-provided repo '{id}'; use 'repo disable' instead"
                    ).format(
                        id=repo_id,
                    )
                )
                return 1

        with ConfigEditor.work_on_user_local_config(cfg) as editor:
            if not editor.remove_repos_entry(repo_id):
                logger.F(
                    _("no repo with id '{id}' found in user config").format(id=repo_id)
                )
                return 1
            editor.stage()

        if args.purge:
            repo_dir = cfg.get_repo_dir_for_id(repo_id)
            if pathlib.Path(repo_dir).exists():
                shutil.rmtree(repo_dir)
                logger.I(_("purged cached data at '{path}'").format(path=repo_dir))

        logger.I(_("repo '{id}' removed").format(id=repo_id))
        return 0


class RepoEnableCommand(
    RepoCommand,
    cmd="enable",
    help=_("Enable a package repository"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        a = p.add_argument("id", type=str, help=_("repository identifier to enable"))
        if gc.is_cli_autocomplete:
            from .cli_completion import repo_id_completer_builder

            a.completer = repo_id_completer_builder(gc)

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from ..config.editor import ConfigEditor
        from ..config.schema import KEY_REPOS_ACTIVE

        logger = cfg.logger
        repo_id: str = args.id

        with ConfigEditor.work_on_user_local_config(cfg) as editor:
            if not editor.update_repos_entry(repo_id, {KEY_REPOS_ACTIVE: True}):
                logger.F(
                    _("no repo with id '{id}' found in user config").format(id=repo_id)
                )
                return 1
            editor.stage()

        logger.I(_("repo '{id}' enabled").format(id=repo_id))
        return 0


class RepoDisableCommand(
    RepoCommand,
    cmd="disable",
    help=_("Disable a package repository"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        a = p.add_argument("id", type=str, help=_("repository identifier to disable"))
        if gc.is_cli_autocomplete:
            from .cli_completion import repo_id_completer_builder

            a.completer = repo_id_completer_builder(gc)

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from ..config.editor import ConfigEditor
        from ..config.schema import KEY_REPOS_ACTIVE

        logger = cfg.logger
        repo_id: str = args.id

        with ConfigEditor.work_on_user_local_config(cfg) as editor:
            if not editor.update_repos_entry(repo_id, {KEY_REPOS_ACTIVE: False}):
                logger.F(
                    _("no repo with id '{id}' found in user config").format(id=repo_id)
                )
                return 1
            editor.stage()

        logger.I(_("repo '{id}' disabled").format(id=repo_id))
        return 0


class RepoSetPriorityCommand(
    RepoCommand,
    cmd="set-priority",
    help=_("Set the priority of a package repository"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        a = p.add_argument("id", type=str, help=_("repository identifier"))
        if gc.is_cli_autocomplete:
            from .cli_completion import repo_id_completer_builder

            a.completer = repo_id_completer_builder(gc)
        p.add_argument("priority", type=int, help=_("new priority value"))

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from ..config.editor import ConfigEditor
        from ..config.schema import KEY_REPOS_PRIORITY

        logger = cfg.logger
        repo_id: str = args.id
        priority: int = args.priority

        with ConfigEditor.work_on_user_local_config(cfg) as editor:
            if not editor.update_repos_entry(repo_id, {KEY_REPOS_PRIORITY: priority}):
                logger.F(
                    _("no repo with id '{id}' found in user config").format(id=repo_id)
                )
                return 1
            editor.stage()

        logger.I(
            _("repo '{id}' priority set to {priority}").format(
                id=repo_id, priority=priority
            )
        )
        return 0



import enum
from typing import Iterable, TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self

from ..config import GlobalConfig
from ..i18n import _
from ..utils.porcelain import PorcelainEntity, PorcelainEntityType
from .distfile import Distfile
from .host import get_native_host
from .list_filter import ListFilter
from .pkg_manifest import BoundPackageManifest, PackageManifestType
from .composite_repo import CompositeRepo


class PkgRemark(enum.StrEnum):
    Latest = "latest"
    LatestPreRelease = "latest-prerelease"
    NoBinaryForCurrentHost = "no-binary-for-current-host"
    PreRelease = "prerelease"
    HasKnownIssue = "known-issue"
    Downloaded = "downloaded"
    Installed = "installed"

    def as_rich_markup(self) -> str:
        match self:
            case self.Latest:
                return _("latest")
            case self.LatestPreRelease:
                return _("latest-prerelease")
            case self.NoBinaryForCurrentHost:
                return _("[red]no binary for current host[/]")
            case self.PreRelease:
                return _("prerelease")
            case self.HasKnownIssue:
                return _("[yellow]has known issue[/]")
            case self.Downloaded:
                return _("[green]downloaded[/]")
            case self.Installed:
                return _("[green]installed[/]")
        return ""


class AugmentedPkgManifest:
    def __init__(
        self,
        pm: BoundPackageManifest,
        remarks: list[PkgRemark],
    ) -> None:
        self.pm = pm
        self.remarks = remarks
        self._is_downloaded = PkgRemark.Downloaded in remarks
        self._is_installed = PkgRemark.Installed in remarks

    def to_porcelain(self) -> "PorcelainPkgVersionV1":
        return {
            "semver": str(self.pm.semver),
            "pm": self.pm.to_raw(),
            "remarks": self.remarks,
            "is_downloaded": self._is_downloaded,
            "is_installed": self._is_installed,
        }


class AugmentedPkg:
    def __init__(self) -> None:
        self.versions: list[AugmentedPkgManifest] = []

    def add_version(self, v: AugmentedPkgManifest) -> None:
        if self.versions:
            if v.pm.category != self.category or v.pm.name != self.name:
                raise ValueError("cannot add a version of a different pkg")
        self.versions.append(v)

    @property
    def category(self) -> str | None:
        return self.versions[0].pm.category if self.versions else None

    @property
    def name(self) -> str | None:
        return self.versions[0].pm.name if self.versions else None

    @classmethod
    def yield_from_repo(
        cls,
        cfg: GlobalConfig,
        mr: CompositeRepo,
        filters: ListFilter,
    ) -> "Iterable[Self]":
        rgs = cfg.ruyipkg_global_state
        native_host = str(get_native_host())

        for category, pkg_name, pkg_vers in mr.iter_pkgs():
            if not filters.check_pkg_name(cfg, mr, category, pkg_name):
                continue

            pkg = cls()

            semvers = [pm.semver for pm in pkg_vers.values()]
            semvers.sort(reverse=True)
            found_latest = False
            for i, sv in enumerate(semvers):
                # TODO: support filter ops against individual versions

                pm = pkg_vers[str(sv)]

                latest = False
                latest_prerelease = i == 0
                prerelease = pm.is_prerelease
                if not found_latest and not prerelease:
                    latest = True
                    found_latest = True

                remarks: list[PkgRemark] = []
                if latest or latest_prerelease or prerelease:
                    if prerelease:
                        remarks.append(PkgRemark.PreRelease)
                    if latest:
                        remarks.append(PkgRemark.Latest)
                    if latest_prerelease and not latest:
                        remarks.append(PkgRemark.LatestPreRelease)
                if pm.service_level.has_known_issues:
                    remarks.append(PkgRemark.HasKnownIssue)
                if bm := pm.binary_metadata:
                    if not bm.is_available_for_current_host:
                        remarks.append(PkgRemark.NoBinaryForCurrentHost)
                if _is_pkg_fully_downloaded(pm):
                    remarks.append(PkgRemark.Downloaded)

                host = native_host if bm is not None else ""
                is_installed = rgs.is_package_installed(
                    pm.repo_id,
                    pm.category,
                    pm.name,
                    str(sv),
                    host,
                )
                if is_installed:
                    remarks.append(PkgRemark.Installed)

                pkg.add_version(AugmentedPkgManifest(pm, remarks))

            yield pkg

    def to_porcelain(self) -> "PorcelainPkgListOutputV1":
        return {
            "ty": PorcelainEntityType.PkgListOutputV1,
            "category": self.category or "",
            "name": self.name or "",
            "vers": [x.to_porcelain() for x in self.versions],
        }


class PorcelainPkgVersionV1(TypedDict):
    semver: str
    pm: PackageManifestType
    remarks: list[PkgRemark]
    is_downloaded: bool
    is_installed: bool


class PorcelainPkgListOutputV1(PorcelainEntity):
    category: str
    name: str
    vers: list[PorcelainPkgVersionV1]


def _is_pkg_fully_downloaded(pm: BoundPackageManifest) -> bool:
    dfs = pm.distfiles
    if not dfs:
        return True

    for df_decl in dfs.values():
        df = Distfile(df_decl, pm.repo)
        if not df.is_downloaded():
            return False

    return True



import datetime
import json
import os
import pathlib
from typing import Any, Iterable, Iterator, TypedDict, TYPE_CHECKING
from dataclasses import dataclass

from .pkg_manifest import BoundPackageManifest
from .protocols import ProvidesPackageManifests

if TYPE_CHECKING:
    # for avoiding heavy import
    from .protocols import ProvidesPackageManifests


class PackageInstallationRecord(TypedDict):
    """Record of a package installation."""

    repo_id: str
    category: str
    name: str
    version: str
    host: str  # For binary packages, empty for blobs
    install_path: str
    install_time: str  # ISO format datetime


@dataclass
class PackageInstallationInfo:
    """Information about an installed package."""

    repo_id: str
    category: str
    name: str
    version: str
    host: str  # For binary packages, empty for blobs
    install_path: str
    install_time: datetime.datetime

    def to_record(self) -> PackageInstallationRecord:
        """Convert to a record for JSON serialization."""
        return PackageInstallationRecord(
            repo_id=self.repo_id,
            category=self.category,
            name=self.name,
            version=self.version,
            host=self.host,
            install_path=self.install_path,
            install_time=self.install_time.isoformat(),
        )

    @classmethod
    def from_record(
        cls,
        record: PackageInstallationRecord,
    ) -> "PackageInstallationInfo":
        """Create from a record."""
        return cls(
            repo_id=record["repo_id"],
            category=record["category"],
            name=record["name"],
            version=record["version"],
            host=record["host"],
            install_path=record["install_path"],
            install_time=datetime.datetime.fromisoformat(record["install_time"]),
        )


class RuyipkgGlobalStateStore:
    def __init__(self, root: os.PathLike[Any]) -> None:
        self.root = pathlib.Path(root)
        self._installs_file = self.root / "installs.json"
        self._installs_cache: dict[str, PackageInstallationInfo] | None = None

    def ensure_state_dir(self) -> None:
        """Ensure the state directory exists."""
        self.root.mkdir(parents=True, exist_ok=True)

    def purge_installation_info(self) -> None:
        """Purge installation records."""
        self._installs_file.unlink(missing_ok=True)
        self._installs_cache = None
        # if the state dir is empty, remove it
        try:
            self.root.rmdir()
        except OSError:
            pass

    def _load_installs(self) -> dict[str, PackageInstallationInfo]:
        """Load installation records from disk."""
        if self._installs_cache is not None:
            return self._installs_cache

        self.ensure_state_dir()

        if not self._installs_file.exists():
            self._installs_cache = {}
            return self._installs_cache

        try:
            with open(self._installs_file, "r", encoding="utf-8") as f:
                data = json.load(f)

            installs = {}
            for key, record in data.items():
                installs[key] = PackageInstallationInfo.from_record(record)

            self._installs_cache = installs
            return self._installs_cache
        except (json.JSONDecodeError, KeyError, ValueError):
            # If file is corrupted, start fresh
            self._installs_cache = {}
            return self._installs_cache

    def _save_installs(self) -> None:
        """Save installation records to disk."""
        if self._installs_cache is None:
            return

        self.ensure_state_dir()

        data = {}
        for key, info in self._installs_cache.items():
            data[key] = info.to_record()

        # Write atomically by writing to temp file then renaming
        temp_file = self._installs_file.with_suffix(".tmp")
        try:
            with open(temp_file, "w", encoding="utf-8") as f:
                json.dump(data, f, indent=2, ensure_ascii=False)
            temp_file.replace(self._installs_file)
        except Exception:
            # Clean up temp file if something went wrong
            if temp_file.exists():
                temp_file.unlink()
            raise

    def _get_installation_key(
        self,
        repo_id: str,
        category: str,
        name: str,
        version: str,
        host: str = "",
    ) -> str:
        """Get the key used to store installation info."""
        if host:
            # Use a format that includes host for binary packages
            return f"{repo_id}:{category}/{name} {version} host={host}"
        return f"{repo_id}:{category}/{name} {version}"

    def record_installation(
        self,
        repo_id: str,
        category: str,
        name: str,
        version: str,
        host: str,
        install_path: str,
    ) -> None:
        """Record a successful package installation."""
        installs = self._load_installs()

        key = self._get_installation_key(repo_id, category, name, version, host)
        info = PackageInstallationInfo(
            repo_id=repo_id,
            category=category,
            name=name,
            version=version,
            host=host,
            install_path=install_path,
            install_time=datetime.datetime.now(),
        )

        installs[key] = info
        self._save_installs()

    def remove_installation(
        self,
        repo_id: str,
        category: str,
        name: str,
        version: str,
        host: str = "",
    ) -> bool:
        """Remove an installation record."""
        installs = self._load_installs()
        key = self._get_installation_key(repo_id, category, name, version, host)

        if key in installs:
            del installs[key]
            self._save_installs()
            return True
        return False

    def get_installation(
        self,
        repo_id: str,
        category: str,
        name: str,
        version: str,
        host: str = "",
    ) -> PackageInstallationInfo | None:
        """Get information about a specific installation."""
        installs = self._load_installs()
        key = self._get_installation_key(repo_id, category, name, version, host)
        return installs.get(key)

    def is_package_installed(
        self,
        repo_id: str,
        category: str,
        name: str,
        version: str,
        host: str = "",
    ) -> bool:
        """Check if a package is installed."""
        return self.get_installation(repo_id, category, name, version, host) is not None

    def list_installed_packages(self) -> list[PackageInstallationInfo]:
        """List all installed packages."""
        installs = self._load_installs()
        return list(installs.values())


class BoundInstallationStateStore(ProvidesPackageManifests):
    def __init__(
        self, rgs: RuyipkgGlobalStateStore, mr: "ProvidesPackageManifests"
    ) -> None:
        self._rgs = rgs
        self._mr = mr

    def _get_installed_manifest(
        self,
        info: PackageInstallationInfo,
    ) -> BoundPackageManifest | None:
        """Get the bound manifest for an installed package, or None if not found in repo."""

        return self._mr.get_pkg(info.name, info.category, info.version)

    def iter_pkg_manifests(self) -> Iterable[BoundPackageManifest]:
        """Iterate over all installed package manifests."""

        installed_pkgs = self._rgs.list_installed_packages()
        for info in installed_pkgs:
            if m := self._get_installed_manifest(info):
                yield m

    def iter_pkgs(
        self,
    ) -> Iterable[tuple[str, str, dict[str, BoundPackageManifest]]]:
        """Iterate over installed packages grouped by category and name."""

        installed_pkgs = self._rgs.list_installed_packages()

        # Group by category and name
        result: dict[str, dict[str, dict[str, BoundPackageManifest]]] = {}
        for info in installed_pkgs:
            if m := self._get_installed_manifest(info):
                if info.category not in result:
                    result[info.category] = {}
                if info.name not in result[info.category]:
                    result[info.category][info.name] = {}
                result[info.category][info.name][info.version] = m

        for category, cat_pkgs in result.items():
            for pkg_name, pkg_vers in cat_pkgs.items():
                yield (category, pkg_name, pkg_vers)

    def iter_pkg_vers(
        self,
        name: str,
        category: str | None = None,
    ) -> Iterable[BoundPackageManifest]:
        """Iterate over installed versions of a specific package."""

        installed_pkgs = self._rgs.list_installed_packages()
        for info in installed_pkgs:
            if info.name == name and (category is None or info.category == category):
                if m := self._get_installed_manifest(info):
                    yield m

    def get_pkg(
        self,
        name: str,
        category: str,
        ver: str,
    ) -> BoundPackageManifest | None:
        """Returns the package manifest by exact match, or None if not found."""
        installed_pkgs = self._rgs.list_installed_packages()
        for info in installed_pkgs:
            if info.name == name and info.category == category and info.version == ver:
                if m := self._get_installed_manifest(info):
                    return m
                # Package is installed but not found in current repo
                break

        return None

    def get_pkg_latest_ver(
        self,
        name: str,
        category: str | None = None,
        include_prerelease_vers: bool = False,
    ) -> BoundPackageManifest:
        """Get the latest installed version of a package."""

        from .pkg_manifest import is_prerelease

        installed_vers = list(self.iter_pkg_vers(name, category))
        if not installed_vers:
            raise KeyError(f"No installed versions found for package '{name}'")

        if not include_prerelease_vers:
            installed_vers = [
                pm for pm in installed_vers if not is_prerelease(pm.semver)
            ]
            if not installed_vers:
                raise KeyError(
                    f"No non-prerelease installed versions found for package '{name}'"
                )

        # Find the latest version
        latest = max(installed_vers, key=lambda pm: pm.semver)
        return latest

    # To be removed later along with slug support
    def get_pkg_by_slug(self, slug: str) -> BoundPackageManifest | None:
        """Get an installed package by its slug."""

        installed_pkgs = self._rgs.list_installed_packages()
        for info in installed_pkgs:
            if m := self._get_installed_manifest(info):
                if m.slug == slug:
                    return m
        return None

    # Useful helpers

    def iter_upgradable_pkgs(
        self,
        include_prereleases: bool = False,
    ) -> Iterator[tuple[BoundPackageManifest, str, bool]]:
        """Iterate over installed packages that have newer versions available.

        Yields ``(installed_pm, new_version_str, repo_migrated)`` tuples.
        ``repo_migrated`` is True when the latest version comes from a
        different repo than the installed version."""
        for installed_pm in self.iter_pkg_manifests():
            latest_pm: BoundPackageManifest
            try:
                latest_pm = self._mr.get_pkg_latest_ver(
                    installed_pm.name,
                    installed_pm.category,
                    include_prereleases,
                )
            except KeyError:
                # package not found in the repo, skip it
                continue

            if latest_pm.semver > installed_pm.semver:
                migrated = latest_pm.repo_id != installed_pm.repo_id
                yield (installed_pm, str(latest_pm.semver), migrated)



import argparse
from typing import TYPE_CHECKING

from ..cli.cmd import RootCommand
from ..i18n import _

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class UpdateCommand(
    RootCommand,
    cmd="update",
    help=_("Update RuyiSDK repo and packages"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        a = p.add_argument(
            "--repo",
            type=str,
            default=None,
            help=_("only sync the repo with this ID"),
        )
        if gc.is_cli_autocomplete:
            from .cli_completion import repo_id_completer_builder

            a.completer = repo_id_completer_builder(gc)

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from . import news
        from .state import BoundInstallationStateStore

        logger = cfg.logger
        mr = cfg.repo

        repo_id: str | None = args.repo
        if repo_id is not None:
            try:
                mr.sync_one(repo_id)
            except ValueError:
                logger.F(_("no active repo with id '{id}'").format(id=repo_id))
                return 1
        else:
            mr.sync_all()

        # check for upgradable packages
        bis = BoundInstallationStateStore(cfg.ruyipkg_global_state, mr)
        upgradable = list(bis.iter_upgradable_pkgs(cfg.include_prereleases))

        if upgradable:
            logger.stdout(
                _(
                    "\nNewer versions are available for some of your installed packages:\n"
                )
            )
            for pm, new_ver, migrated in upgradable:
                logger.stdout(
                    f"  - [bold]{pm.category}/{pm.name}[/]: [yellow]{pm.ver}[/] -> [green]{new_ver}[/]"
                )
                if migrated:
                    logger.W(
                        _(
                            "package '{category}/{name}' was installed from "
                            "repo '{repo}' but the latest version is in a "
                            "different repo"
                        ).format(
                            category=pm.category,
                            name=pm.name,
                            repo=pm.repo_id,
                        )
                    )
            logger.stdout(
                _(
                    """
Re-run [yellow]ruyi install[/] to upgrade, and don't forget to re-create any affected
virtual environments."""
                )
            )

        news.maybe_notify_unread_news(cfg, False)

        return 0



import pathlib
import tomllib

import pytest

from ruyi.config.editor import ConfigEditor
from ruyi.config.errors import (
    InvalidConfigKeyError,
    InvalidConfigSectionError,
    InvalidConfigValueTypeError,
    MalformedConfigFileError,
    ProtectedGlobalConfigError,
)


@pytest.fixture
def temp_config_file(tmp_path: pathlib.Path) -> pathlib.Path:
    return tmp_path / "config.toml"


def test_enter_exit(temp_config_file: pathlib.Path) -> None:
    editor = ConfigEditor(temp_config_file)
    with editor as e:
        assert e is editor
        e.set_value("telemetry.mode", "off")
    # no stage() so no file writing
    assert not temp_config_file.exists()


def test_set_value(temp_config_file: pathlib.Path) -> None:
    with ConfigEditor(temp_config_file) as e:
        with pytest.raises(InvalidConfigKeyError):
            e.set_value("invalid_key", "value")

        with pytest.raises(InvalidConfigValueTypeError):
            e.set_value("telemetry.mode", True)

        with pytest.raises(InvalidConfigValueTypeError):
            e.set_value("telemetry.mode", 1)

        with pytest.raises(ProtectedGlobalConfigError):
            e.set_value("installation.externally_managed", True)

        e.set_value("telemetry.mode", "off")
        e.stage()

    with open(temp_config_file, "rb") as fp:
        content = tomllib.load(fp)
    assert "installation" not in content
    assert content["telemetry"]["mode"] == "off"


def test_unset_value_remove_section(temp_config_file: pathlib.Path) -> None:
    with ConfigEditor(temp_config_file) as e:
        e.set_value("telemetry.mode", "off")
        e.set_value("repo.remote", "http://test.example.com")
        e.stage()

    with open(temp_config_file, "rb") as fp:
        content = tomllib.load(fp)
    assert content["repo"]["remote"] == "http://test.example.com"
    assert content["telemetry"]["mode"] == "off"

    with ConfigEditor(temp_config_file) as e:
        e.unset_value("telemetry.mode")
        e.remove_section("repo")
        e.stage()

        with pytest.raises(InvalidConfigSectionError):
            e.remove_section("foo")

    with open(temp_config_file, "rb") as fp:
        content = tomllib.load(fp)
    assert "repo" not in content
    assert "telemetry" in content
    assert "mode" not in content["telemetry"]


def test_malformed_config_file_error(temp_config_file: pathlib.Path) -> None:
    with open(temp_config_file, "wb") as fp:
        fp.write(b"repo = 1\n")

    with pytest.raises(MalformedConfigFileError):
        with ConfigEditor(temp_config_file) as e:
            e.set_value("repo.branch", "foo")



"""Tests for the recipe build runner (B7)."""

from __future__ import annotations

import pathlib

import pytest

from ruyi.log import RuyiLogger
from ruyi.ruyipkg.build_runner import (
    BuildFailure,
    format_build_report,
    run_recipe,
)


def _make_project(tmp_path: pathlib.Path, recipe_body: str) -> pathlib.Path:
    (tmp_path / "ruyi-build-recipes.toml").write_text(
        'format = "v1"\n[project]\nname = "testproj"\n'
    )
    (tmp_path / "out").mkdir(exist_ok=True)
    recipe_dir = tmp_path / "recipes"
    recipe_dir.mkdir(exist_ok=True)
    recipe = recipe_dir / "pkg.star"
    recipe.write_text(recipe_body)
    return recipe


def test_run_recipe_dry_run(tmp_path: pathlib.Path, ruyi_logger: RuyiLogger) -> None:
    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/true', 'hi'])\n"
        "\n"
        "RUYI.build.schedule_build(build_it)\n",
    )

    reports = run_recipe(ruyi_logger, recipe, dry_run=True)
    assert len(reports) == 1
    r = reports[0]
    assert r.build_name == "build_it"
    assert r.invocations[0].argv == ("/bin/true", "hi")
    assert r.exit_code == 0
    assert r.artifacts == ()


def test_run_recipe_no_scheduled_builds_errors(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    recipe = _make_project(tmp_path, "RUYI = ruyi_plugin_rev(1)\n")
    with pytest.raises(RuntimeError, match="scheduled no builds"):
        run_recipe(ruyi_logger, recipe, dry_run=True)


def test_run_recipe_name_filter_selects(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_a(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/true', 'a'])\n"
        "def build_b(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/true', 'b'])\n"
        "\n"
        "RUYI.build.schedule_build(build_a)\n"
        "RUYI.build.schedule_build(build_b)\n",
    )
    reports = run_recipe(ruyi_logger, recipe, dry_run=True, selected_names=["build_b"])
    assert [r.build_name for r in reports] == ["build_b"]


def test_run_recipe_name_filter_missing_errors(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_a(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/true'])\n"
        "RUYI.build.schedule_build(build_a)\n",
    )
    with pytest.raises(RuntimeError, match="does not define the requested"):
        run_recipe(ruyi_logger, recipe, dry_run=True, selected_names=["nope"])


def test_run_recipe_user_vars(tmp_path: pathlib.Path, ruyi_logger: RuyiLogger) -> None:
    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    arch = ctx.var('arch')\n"
        "    return ctx.subprocess(argv = ['/bin/echo', arch])\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    reports = run_recipe(
        ruyi_logger,
        recipe,
        dry_run=True,
        user_vars={"arch": "riscv64"},
    )
    assert reports[0].invocations[0].argv == ("/bin/echo", "riscv64")


def test_run_recipe_executes_and_collects_artifacts(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    out = tmp_path / "out"
    out.mkdir(exist_ok=True)
    artifact = out / "pkg-1.0.tar.zst"
    artifact.write_bytes(b"dummy")

    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(\n"
        "        argv = ['/bin/true'],\n"
        "        produces = [ctx.artifact(glob = 'pkg-*.tar.zst')],\n"
        "    )\n"
        "RUYI.build.schedule_build(build_it)\n",
    )

    reports = run_recipe(ruyi_logger, recipe)
    assert len(reports) == 1
    r = reports[0]
    assert r.exit_code == 0
    assert len(r.artifacts) == 1
    ar = r.artifacts[0]
    assert ar.path == artifact
    assert ar.size == 5
    assert len(ar.checksums["sha256"]) == 64
    assert len(ar.checksums["sha512"]) == 128


def test_run_recipe_missing_artifact_fails(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(\n"
        "        argv = ['/bin/true'],\n"
        "        produces = [ctx.artifact(glob = 'nope-*.tar')],\n"
        "    )\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    with pytest.raises(RuntimeError, match="matched no files"):
        run_recipe(ruyi_logger, recipe)


def test_run_recipe_non_zero_exit_raises_build_failure(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/false'])\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    with pytest.raises(BuildFailure) as excinfo:
        run_recipe(ruyi_logger, recipe)
    assert excinfo.value.exit_code != 0
    assert excinfo.value.build_name == "build_it"


def test_run_recipe_supports_ruyi_build_load(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    (tmp_path / "ruyi-build-recipes.toml").write_text('format = "v1"\n')
    (tmp_path / "out").mkdir(exist_ok=True)
    lib = tmp_path / "lib"
    lib.mkdir()
    (lib / "common.star").write_text("GREETING = 'hi from lib'\n")

    recipe = tmp_path / "recipes" / "pkg.star"
    recipe.parent.mkdir()
    recipe.write_text(
        "RUYI = ruyi_plugin_rev(1)\n"
        "load('ruyi-build://lib/common.star', 'GREETING')\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/echo', GREETING])\n"
        "RUYI.build.schedule_build(build_it)\n"
    )

    reports = run_recipe(ruyi_logger, recipe, dry_run=True)
    assert reports[0].invocations[0].argv == ("/bin/echo", "hi from lib")


def test_format_build_report_is_reasonable(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(argv = ['/bin/true'])\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    reports = run_recipe(ruyi_logger, recipe, dry_run=True)
    text = format_build_report(reports[0])
    assert 'build_name = "build_it"' in text
    assert "[[invocations]]" in text


def test_format_build_report_includes_artifacts(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    out = tmp_path / "out"
    out.mkdir(exist_ok=True)
    (out / "pkg-1.0.tar.zst").write_bytes(b"dummy")

    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(\n"
        "        argv = ['/bin/true'],\n"
        "        produces = [ctx.artifact(glob = 'pkg-*.tar.zst')],\n"
        "    )\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    reports = run_recipe(ruyi_logger, recipe)
    text = format_build_report(reports[0])
    ar = reports[0].artifacts[0]
    assert "[[artifacts]]" in text
    assert f'path = "{ar.path}"' in text
    assert f"size = {ar.size}" in text
    assert f'sha256 = "{ar.checksums["sha256"]}"' in text
    assert f'sha512 = "{ar.checksums["sha512"]}"' in text


def test_run_recipe_rejects_non_invocation_return(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return 'not-an-invocation'\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    with pytest.raises(RuntimeError, match="expected Invocation"):
        run_recipe(ruyi_logger, recipe, dry_run=True)


def test_run_recipe_rejects_list_with_non_invocation(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return [ctx.subprocess(argv = ['/bin/true']), 42]\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    with pytest.raises(RuntimeError, match="expected Invocation"):
        run_recipe(ruyi_logger, recipe, dry_run=True)


def test_run_recipe_rejects_empty_list(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return []\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    with pytest.raises(RuntimeError, match="empty list"):
        run_recipe(ruyi_logger, recipe, dry_run=True)


def test_run_recipe_output_dir_override(
    tmp_path: pathlib.Path, ruyi_logger: RuyiLogger
) -> None:
    override_out = tmp_path / "override_out"
    override_out.mkdir()
    (override_out / "overridden.tar.zst").write_bytes(b"ok")

    recipe = _make_project(
        tmp_path,
        "RUYI = ruyi_plugin_rev(1)\n"
        "def build_it(ctx):\n"
        "    return ctx.subprocess(\n"
        "        argv = ['/bin/true'],\n"
        "        produces = [ctx.artifact(glob = '*.tar.zst')],\n"
        "    )\n"
        "RUYI.build.schedule_build(build_it)\n",
    )
    reports = run_recipe(ruyi_logger, recipe, output_dir_override=override_out)
    assert len(reports) == 1
    assert len(reports[0].artifacts) == 1
    assert (
        reports[0].artifacts[0].path == (override_out / "overridden.tar.zst").resolve()
    )



import shlex
from typing import Any, Final, Callable, Tuple

from jinja2 import BaseLoader, Environment, TemplateNotFound

from ..resource_bundle import get_template_str


class EmbeddedLoader(BaseLoader):
    def __init__(self) -> None:
        pass

    def get_source(
        self,
        environment: Environment,
        template: str,
    ) -> Tuple[str, str | None, Callable[[], bool] | None]:
        if payload := get_template_str(template):
            return payload, None, None
        raise TemplateNotFound(template)


_JINJA_ENV: Final = Environment(
    loader=EmbeddedLoader(),
    autoescape=False,  # we're not producing HTML
    auto_reload=False,  # we're serving statically embedded assets
    keep_trailing_newline=True,  # to make shells happy
)
_JINJA_ENV.filters["sh"] = shlex.quote


def render_template_str(template_name: str, data: dict[str, Any]) -> str:
    tmpl = _JINJA_ENV.get_template(template_name)
    return tmpl.render(data)



import re

SPLIT_RE: re.Pattern[str]

class JsonSchemaException(ValueError): ...

class JsonSchemaValueException(JsonSchemaException):
    message: str
    value: object | None
    name: str | None
    definition: object | None
    rule: str | None
    def __init__(
        self,
        message: str,
        value: object | None = None,
        name: str | None = None,
        definition: object | None = None,
        rule: str | None = None,
    ) -> None: ...
    @property
    def path(self) -> list[str]: ...
    @property
    def rule_definition(self) -> object: ...

class JsonSchemaDefinitionException(JsonSchemaException): ...



import glob
import os
import pathlib
import platform
import re
import subprocess
import sys
from typing import Final, Mapping, TypedDict, TYPE_CHECKING
import uuid

if TYPE_CHECKING:
    from typing_extensions import NotRequired

from .ci import probe_for_ci


class NodeInfo(TypedDict):
    v: int
    report_uuid: str

    arch: str
    ci: str
    libc_name: str
    libc_ver: str
    os: str
    os_release_id: str
    os_release_version_id: str
    shell: str

    riscv_machine: "NotRequired[RISCVMachineInfo]"


class RISCVMachineInfo(TypedDict):
    model_name: str
    cpu_count: int
    isa: str
    uarch: str
    uarch_csr: str
    mmu: str


def probe_for_libc() -> tuple[str, str]:
    r = platform.libc_ver()
    if r[0] and r[1]:
        return r

    # check for musl ld.so at the upstream standard paths, because
    # platform.libc_ver() as of Python 3.12 does not know how to handle musl
    #
    # see https://wiki.musl-libc.org/guidelines-for-distributions
    musl_lds = glob.glob("/lib/ld-musl-*.so.1")
    if musl_lds:
        # run it and check for "Version *.*.*"
        # in case of multiple hits (hybrid-architecture sysroot?), hope the
        # first one that successfully returns something is the native one
        for p in musl_lds:
            if ver := _try_get_musl_ver(p):
                return ("musl", ver)

    return ("unknown", "unknown")


_MUSL_VERSION_RE: Final = re.compile(rb"(?m)^Version ([0-9.]+)$")


def _try_get_musl_ver(ldso_path: str) -> str | None:
    res = subprocess.run([ldso_path], stderr=subprocess.PIPE)
    if m := _MUSL_VERSION_RE.search(res.stderr):
        return m.group(1).decode("ascii", "ignore")
    return None


def _try_parse_hex(v: str) -> int | None:
    if not v.startswith("0x"):
        return None
    try:
        return int(v[2:], 16)
    except ValueError:
        return None


def probe_for_riscv_machine_info(
    model_name: str | None = None,
    cpuinfo_data: str | None = None,
) -> RISCVMachineInfo | None:
    if model_name is None:
        try:
            with open(
                "/sys/firmware/devicetree/base/model",
                "r",
                encoding="utf-8",
            ) as fp:
                model_name = fp.read().strip(" \n\t\x00")
        except Exception:
            pass

        if not model_name:
            model_name = "unknown"

    if cpuinfo_data is None:
        try:
            with open("/proc/cpuinfo", "r", encoding="utf-8") as fp:
                cpuinfo_data = fp.read()
        except Exception:
            pass

    cpu_count = 0
    isa, mmu, uarch = "unknown", "unknown", "unknown"
    mvendorid: int | None = None
    marchid: int | None = None
    mimpid: int | None = None
    if cpuinfo_data is not None:
        for line in cpuinfo_data.split("\n"):
            if not line:
                continue

            try:
                k, v = line.split(": ", 1)
            except ValueError:
                # malformed line: non-empty but no ": "
                continue

            k = k.strip(" \t")
            v = v.strip()

            match k:
                case "processor":
                    cpu_count += 1
                case "isa":
                    isa = v
                case "mmu":
                    mmu = v
                case "uarch":
                    uarch = v
                case "mvendorid":
                    mvendorid = _try_parse_hex(v)
                case "marchid":
                    marchid = _try_parse_hex(v)
                case "mimpid":
                    mimpid = _try_parse_hex(v)
                case _:
                    continue

    if mvendorid is not None and marchid is not None and mimpid is not None:
        uarch_csr = f"{mvendorid:x}:{marchid:x}:{mimpid:x}"
    else:
        uarch_csr = "unknown"

    return {
        "model_name": model_name,
        "cpu_count": cpu_count,
        "isa": isa,
        "mmu": mmu,
        "uarch": uarch,
        "uarch_csr": uarch_csr,
    }


def probe_for_shell(os_environ: Mapping[str, str]) -> str:
    if x := os_environ.get("SHELL"):
        return os.path.basename(x)
    return "unknown"


def probe_for_container_runtime(os_environ: Mapping[str, str]) -> str:
    """Check if we are likely running in a container. Probes FS and environment
    for signatures of known container runtimes."""

    # check environment markers first

    if "KUBERNETES_SERVICE_HOST" in os_environ:
        return "kubernetes"

    if "container" in os_environ:
        v = os_environ["container"].lower()
        if v == "oci":
            return "other-oci-compliant"
        # could be e.g. "lxc", "lxc-libvirt", "systemd-nspawn", etc.
        return v

    # check for filesystem markers

    if os.path.exists("/run/.containerenv"):
        return "podman"
    # Docker must be checked after Podman
    if os.path.exists("/.dockerenv"):
        return "docker"

    try:
        v = pathlib.Path("/run/systemd/container").read_text(encoding="utf-8").strip()
        if v:
            return v.lower()
    except Exception:
        pass

    if _probe_for_wsl():
        return "wsl"

    return "unknown"


def _probe_for_wsl() -> bool:
    if sys.platform != "linux":
        return False
    # http://github.com/Microsoft/WSL/issues/423#issuecomment-221627364
    for path in ("/proc/sys/kernel/osrelease", "/proc/version"):
        try:
            ver = pathlib.Path(path).read_text(encoding="utf-8")
        except Exception:
            continue
        return "Microsoft" in ver or "WSL" in ver
    return False


def gather_node_info(report_uuid: uuid.UUID | None = None) -> NodeInfo:
    arch = platform.machine()
    libc = probe_for_libc()
    os_release = platform.freedesktop_os_release()

    os_version = os_release.get("VERSION_CODENAME")  # works on e.g. Debian
    if not os_version:
        os_version = os_release.get("VERSION_ID")  # works on e.g. openEuler, Gentoo
    if not os_version:
        os_version = "unknown"

    data: NodeInfo = {
        "v": 1,
        "report_uuid": report_uuid.hex if report_uuid is not None else uuid.uuid4().hex,
        "arch": arch,
        "ci": probe_for_ci(os.environ) or "maybe-not",
        "libc_name": libc[0],
        "libc_ver": libc[1],
        "os": sys.platform,
        "os_release_id": os_release.get("ID", "unknown"),
        "os_release_version_id": os_version,
        "shell": probe_for_shell(os.environ),
    }

    if arch.startswith("riscv"):
        if riscv_machine := probe_for_riscv_machine_info():
            data["riscv_machine"] = riscv_machine

    return data
